import java.util.Arrays;
import java.io.*;

public class IMDDiskImage implements DiskImage {
	private byte[] img = null;
	private String title;
	private int[] sectbl;
	private int nsec;
	private int sec0;
	private int nbyt;
	private int lat;
	private int ntrk;
	private int nsid;
	private boolean density;
	private boolean interlaced;
	private byte[] sb = null;
	private int media; // 5 or 8

	public IMDDiskImage(File f, boolean intl) {
		byte[] imd;
		try {
			imd = new byte[(int)f.length()];
			FileInputStream fi = new FileInputStream(f);
			fi.read(imd);
			fi.close();
			if (imd.length < 4 || imd[0] != 'I' || imd[1] != 'M' ||
					imd[2] != 'D' || imd[3] != ' ') {
				return;
			}
		} catch (Exception ee) {
			return;
		}
		nsec = 0;
		if (!check(imd)) {
			return;
		}
		if (ntrk == 77) {
			media = 8;
		} else if (ntrk == 40 || ntrk == 80) {
			media = 5;
		} else {
			return;
		}
		img = imd;
		// TODO: need to know "interlaced" format...
		interlaced = intl;
	}

	public boolean isError() {
		return img == null;
	}

	public int length() {
		return img == null ? 0 : img.length;
	}

	private int getTrack(byte[] b, int start, RandomAccessFile f) {
		int x = 0;
		boolean dens = (b[start + x++] & 0xff) >= 3;
		int trk = b[start + x++] & 0xff;
		int sid = b[start + x++] & 0xff;
		boolean cylmap = (sid & 0x80) != 0;
		boolean hedmap = (sid & 0x40) != 0;
		sid &= 1;
		int spt = b[start + x++] & 0xff;
		int ssz = (128 << (b[start + x++] & 0x03));
		if (nsec == 0) {
			density = dens;
			nsec = spt;
			nbyt = ssz;
			sectbl = new int[nsec];
			int min = nsec;
			for (int z = 0; z < nsec; ++z) {
				sectbl[z] = b[start + x++] & 0xff;
				if (sectbl[z] < min) {
					min = sectbl[z];
				}
			}
			lat = sectbl[1] - sectbl[0];
			sec0 = min;
		} else {
			if (nsec != spt || nbyt != ssz || density != dens) {
				System.err.format("%d,%d SPT/SSZ mismatch in IMD file " +
					"(%d:%d %d:%d %s:%s)\n", trk, sid,
					nsec, spt, nbyt, ssz, density, dens);
// Some corrupted Kaypro images have garbage at the end...
// if (trk >= 40) { return 99999999; }
				return -1;
			}
			x += spt; // skip sector table
		}
		if (cylmap) {
			x += spt; // skip cylinder table
		}
		if (hedmap) {
			x += spt; // skip head table
		}
		if (f == null) {
			if (ntrk > 0 && trk != ntrk) {
				// Too noisy for interlaced images
				//System.err.format("Warning: track %d:%d out of sequence (%d)\n", trk, sid, ntrk);
			}
			if (trk >= ntrk) ntrk = trk + 1;
			if (sid >= nsid) nsid = sid + 1;
		}
		int trkoff = 0;
		if (interlaced) {
			trkoff = (trk * nsid + sid) * spt * ssz;
		} else {
			trkoff = (sid * ntrk + trk) * spt * ssz;
		}
		for (int z = 0; z < spt; ++z) {
			int sec = sectbl[z];
			int off = trkoff + (sec - sec0) * ssz;
			int type = b[start + x++] & 0xff;
			--type;
			if ((type & 0x7e) != 0) {
				System.err.format("Warning: Unsupported sector type %02x " +
					"in IMD file at %d %d %d (%d)\n",
					type, sid, trk, sec, z);
				type &= 1;
			}
			if (type == 0) {
				if (f != null) try {
					f.seek(off);
					f.write(b, start + x, ssz);
				} catch (Exception ee) {
					return -1;
				}
				x += ssz;
			} else {
				if (f != null) try {
					if (sb == null) {
						sb = new byte[ssz];
					}
					Arrays.fill(sb, (byte)(b[start + x] & 0xff));
					f.seek(off);
					f.write(sb, 0, ssz);
				} catch (Exception ee) {
					return -1;
				}
				++x;
			}
		}
		trkoff += (spt * ssz);
		return x;
	}

	// Scan IMD file for errors, etc.
	private boolean check(byte[] imd) {
		int x = 0;
		while (x < imd.length && (imd[x] & 0xff) != 0x1a) {
			++x;
		}
		if (x >= imd.length) {
			System.err.format("No header terminator - bad IMD file\n");
			return false;
		}
		title = new String(imd, 0, x);
		++x; // skip ^Z
		do {
			int n = getTrack(imd, x, null);
			if (n < 0) {
				return false;
			}
			x += n;
		} while (x < imd.length);
		return true;
	}

	public boolean overlay(File f) {
		if (img == null) {
			return false;
		}
		RandomAccessFile fi = null;
		try {
			fi = new RandomAccessFile(f, "rw");
		} catch (Exception ee) {
			return false;
		}
		int x = 0;
		while (x < img.length && (img[x] & 0xff) != 0x1a) {
			++x;
		}
		++x;
		do {
			int n = getTrack(img, x, fi);
			if (n < 0) {
				// should never happen on this pass...
				return false;
			}
			x += n;
		} while (x < img.length);
		try {
			fi.close();
		} catch (Exception ee) {
			return false;
		}
		return true;
	}

	public int mediaSize() { return media; }
	public int numSides() { return nsid; }
	public int numTracks() { return ntrk; }
	public int sectorsPerTrack() { return nsec; }
	public int sectorSize() { return nbyt; }
	public boolean interlaced() { return interlaced; }
	public int capacity() { return nsid * ntrk * nsec * nbyt; }
	public String getTitle() { return title; }
}
