#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#ifdef CPMTOOLS
#include <utime.h>
#include "cpmfs.h"
#endif

// Program to examine disk images from format.jar or fmtsasi.jar and
// generate a diskdef for use with cpmtools. Uses the MMS-style
// partition table for fmtsasi.jar images.
//
// NOTE: Kaypro uses the winchester disk as two partitions
// with every other pair of tracks belonging to a given partition.
// This cannot be represented to cpmtools (AFAIK).

// Command syntax: getdiskdef <image-file> [<partition-num>]

static int mediaSize_m;
static int secSize_m;
static int numSectors_m;
static int numSides_m;
static int numTracks_m;
static int doubleDensity_m;
static int interlaced_m;
static int mediaLat_m;
static int mediaCyl;
static int mediaHead;
static int mediaSec_m = 0;	// i.e. hard-sectored count
static int result;

static int checkHeader(char *buf, int n) {
    char *b = buf;
    result = 0;

    while (*b != '\n' && *b != '\0' && b - buf < n) {
        char *e;
        int p = strtoul(b, &e, 0);

        switch (tolower(*e)) {
        case 'm':
            if (p != 5 && p != 8) {
                break;
            }
            result |= 0x01;
            mediaSize_m = p;
            break;

        case 'z':
            if (p != 128 && p != 256 && p != 512 && p != 1024) {
                break;
            }
            result |= 0x02;
            secSize_m = p;
            break;

        case 'p':
            if (p == 0) {
                break;
            }
            result |= 0x04;
            numSectors_m = p;
            break;

        case 's':
            if (p < 1 || p > 2) {
                break;
            }
            result |= 0x08;
            numSides_m = p;
            break;

        case 't':
            if (p == 0) {
                break;
            }
            result |= 0x10;
            numTracks_m = p;
            break;

        case 'd':
            result |= 0x20;
            doubleDensity_m = p;
            break;

        case 'i':
            result |= 0x40;
            interlaced_m = p;
            break;

        case 'c':
            if (p == 0) {
                break;
            }
            result |= 0x80;
            mediaCyl = p;
            break;

        case 'h': /* might be SASI or Z17... */
	    if (result & 0x01) {
            	mediaSec_m = p;
                break;
	    }
            if (p == 0) {
                break;
            }
            result |= 0x0100;
            mediaHead = p;
            break;

        case 'l': // optional ?
            mediaLat_m = p;
            break;

        default:
            return 0;
        }

        b = e + 1;
    }

    return (result == 0x7f || result == 0x186);
}

struct dpb {
	unsigned short spt;
	unsigned char bsh, bsm, exm;
	unsigned short dsm, drm;
	unsigned char alv0, alv1;
	unsigned short cks;
	unsigned short restrk;
	unsigned char pad[6]; // non-CP/M stuff (MMS only)
} __attribute__ ((packed));

struct sasi_fdisk {
	unsigned char jmp[3];
	unsigned char drive;
	unsigned char ctrl;
	unsigned char dcharistics[8];
	unsigned char assdtype[6];
	unsigned char nparts;
	unsigned char parts[9][3];
	struct dpb dpbs[8];
} __attribute__ ((packed));

#ifndef CPMTOOLS
// From cpmtools...
struct cpmSuperBlock {
	// struct Device dev;
	struct Device {
		// ...
		int fd;
	} dev;
	// int uppercase;
	int secLength;
	int tracks;
	int sectrk;
	int blksiz;
	int maxdir;
	int dirblks;
	int skew;
	int boottrk;	// added to offset...
	off_t offset;	// added to boottrk...
	int size;
	int extents;
	// ...
};
#endif

static float check_os(struct cpmSuperBlock *ddef) {
	float os = 2.2;
	unsigned char buf[128];
	long off = ddef->offset;
	off += ddef->boottrk * ddef->sectrk * ddef->secLength;
	lseek(ddef->dev.fd, off, SEEK_SET);
	size_t rc = read(ddef->dev.fd, buf, sizeof(buf));
	if (rc != sizeof(buf)) {
		return os;	// error, really
	}
	// Without a disk label, all are "2.2" filesystems... ?
	// In the absence of a label, XFCBs would be ignored.
	// However, cpmtools would see "os 2.2" and assume the
	// XFCBs are just files in user nums 16-31 (and erroneously
	// interpret password/timestamps as allocation data).
	// This could only happen if the directory label were
	// removed later, and the XFCBs left behind.
	if (buf[96] == 0x21) {
		// CP/M Plus timestamp SFCBs present.
		os = 3.1;
	} else if (buf[0] == 0x20 && (buf[12] & 0x70)) {
		// not strickly exclusive to MP/M, but in absence of
		// CP/M Plus SFCBs this must be MP/M XFCBs for timestamps.
		os = 3.0;
	}
	return os;
}

static int count_ones(unsigned char alv0, unsigned char alv1) {
	unsigned int alv = (alv0 << 8) | alv1;
	int x;
	int n = 0;
	for (x = 0; x < 16; ++x) {
		if (alv & 1) ++n;
		alv >>= 1;
	}
	return n;
}

static void set_diskdef(struct dpb *dpb, int off, int ssz, int lat, struct cpmSuperBlock *ddef) {
	ddef->blksiz = (128 << dpb->bsh);
	ddef->secLength = ssz;
	ddef->sectrk = dpb->spt / (ssz / 128);
	ddef->extents = dpb->exm + 1;
	int trksize = (dpb->spt * 128);
	ddef->tracks = ((dpb->dsm + 1) * ddef->blksiz + trksize - 1) / trksize + dpb->restrk;
	ddef->maxdir = dpb->drm + 1;
	ddef->dirblks = count_ones(dpb->alv0, dpb->alv1);
	ddef->skew = lat ? lat : 1;
	ddef->boottrk = dpb->restrk;
	ddef->offset = off;
	ddef->size = (ddef->secLength * ddef->sectrk * (ddef->tracks - ddef->boottrk)) /
							ddef->blksiz;
#ifdef CPMTOOLS
	float os = check_os(ddef);
	if (os < 3.0) {
		ddef->type |= CPMFS_DR22;
	} else if (os == 3.0) {	// MP/M Uses BDOS 3.0, CP/M Plus uses 3.1
		ddef->type |= CPMFS_MPM;
	} else {
		ddef->type |= CPMFS_DR3;
	}
#endif
}

static void get_sasi_diskdef(struct cpmSuperBlock *ddef, void *ptab, int partno) {
	struct sasi_fdisk *fdisk = (struct sasi_fdisk *)ptab;
#if 0
	{
	int x;
	for (x = 0; x < fdisk->nparts; ++x) {
		int y = (fdisk->parts[x][0] << 16) |
			(fdisk->parts[x][1] << 8) |
			fdisk->parts[x][2];
		printf("%d: %d\n", x, y);
	}
	for (x = 0; x < fdisk->nparts; ++x) {
		printf("%d dpb: %d %d %d %d %d %d %02x%02x %d %d\n", x,
			fdisk->dpbs[x].spt,
			fdisk->dpbs[x].bsh, fdisk->dpbs[x].bsm, fdisk->dpbs[x].exm,
			fdisk->dpbs[x].dsm, fdisk->dpbs[x].drm,
			fdisk->dpbs[x].alv0, fdisk->dpbs[x].alv1,
			fdisk->dpbs[x].cks,
			fdisk->dpbs[x].restrk);
	}
	}
#endif
	if (partno >= fdisk->nparts) {
		// TODO: error
		return;
	}
	int offset = (fdisk->parts[partno][0] << 16) |
			(fdisk->parts[partno][1] << 8) |
			fdisk->parts[partno][2];
	int sxf = secSize_m / 128;
	set_diskdef(&fdisk->dpbs[partno], offset * 128, secSize_m, 1, ddef);
}

enum fmts { MMS, Z17, M47, Z37, Z47, Z67, Z37X, Z47X, KAYPRO, HIDENS, FMTMAX };
enum medias { F525, F8, MEDMAX };
enum densities { SD, DD, QD, DENSMAX };
enum sides { SS, DS, SIDSMAX };
enum tpis { ST, DT, QT, TPISMAX };

// pad[0] is software sector skew (latency) for disk format.
// This is used to eliminate skew from formats that show
// "hard skew" i.e. the skew used during formatting.
// TODO: need to fix inconsistency - it may have just been
// early versions of H89FloppyFormatter that did that.
static struct dpb floppy_dpbs[MEDMAX][FMTMAX][DENSMAX][SIDSMAX][TPISMAX] = {
[F525][Z17][SD][SS][ST] = { 20, 3,  7, 0,  91,  63, 0xc0, 0x00, 16, 3,
			{4, 0, 0, 0, 0, 0}},
[F525][Z17][SD][DS][ST] = { 20, 3,  7, 0, 181,  63, 0xc0, 0x00, 16, 3,
			{4, 0, 0, 0, 0, 0}},
[F525][Z17][SD][SS][DT] = { 20, 4, 15, 1,  95, 127, 0xc0, 0x00, 32, 3,
			{4, 0, 0, 0, 0, 0}},
[F525][Z17][SD][DS][DT] = { 20, 4, 15, 1, 185, 127, 0xc0, 0x00, 32, 3,
			{4, 0, 0, 0, 0, 0}},

[F8][M47][DD][SS][ST] = { 64, 4, 15, 0, 299, 191, 0xe0, 0x00, 48, 2,
			{4, 0, 0, 0, 0, 0}},
[F8][M47][DD][DS][ST] = { 64, 4, 15, 0, 607, 191, 0xe0, 0x00, 48, 2,
			{4, 0, 0, 0, 0, 0}},

[F525][MMS][DD][SS][ST] = { 36, 4, 15, 1,  82,  95, 0xc0, 0x00, 24, 3,
			{0, 0, 0, 0, 0, 0}},
[F525][MMS][DD][DS][ST] = { 36, 4, 15, 1, 172,  95, 0xc0, 0x00, 24, 3,
			{0, 0, 0, 0, 0, 0}},
[F525][MMS][DD][SS][DT] = { 36, 5, 31, 3,  85, 127, 0x80, 0x00, 32, 3,
			{0, 0, 0, 0, 0, 0}},
[F525][MMS][DD][DS][DT] = { 36, 5, 31, 3, 175, 127, 0x80, 0x00, 32, 3,
			{0, 0, 0, 0, 0, 0}},
[F8][MMS][DD][SS][ST]   = { 64, 4, 15, 0, 299, 191, 0xe0, 0x00, 48, 2,
			{0, 0, 0, 0, 0, 0}},
[F8][MMS][DD][DS][ST]   = { 64, 4, 15, 0, 607, 191, 0xe0, 0x00, 48, 2,
			{0, 0, 0, 0, 0, 0}},

[F525][Z37][SD][SS][ST] = { 20, 3,  7, 0,  91,  63, 0xc0, 0x00, 16, 3,
			{0, 0, 0, 0, 0, 0}},
[F525][Z37][SD][DS][ST] = { 20, 3,  7, 0, 187, 127, 0xf0, 0x00, 32, 3,
			{0, 0, 0, 0, 0, 0}},
[F525][Z37][SD][SS][DT] = { 20, 3,  7, 0, 191,  63, 0xc0, 0x00, 16, 3,
			{0, 0, 0, 0, 0, 0}},
[F525][Z37][SD][DS][DT] = { 20, 4, 15, 1, 195, 127, 0xc0, 0x00, 32, 3,
			{0, 0, 0, 0, 0, 0}},
[F525][Z37][DD][SS][ST] = { 32, 3,  7, 0, 151, 127, 0xf0, 0x00, 32, 2,
			{0, 0, 0, 0, 0, 0}},
[F525][Z37][DD][DS][ST] = { 32, 4, 15, 0, 155, 255, 0xf0, 0x00, 64, 2,
			{0, 0, 0, 0, 0, 0}},
[F525][Z37][DD][SS][DT] = { 32, 4, 15, 1, 155, 127, 0xc0, 0x00, 32, 2,
			{0, 0, 0, 0, 0, 0}},
[F525][Z37][DD][DS][DT] = { 32, 4, 15, 0, 315, 255, 0xf0, 0x00, 64, 2,
			{0, 0, 0, 0, 0, 0}},

[F525][Z37X][DD][SS][ST] = { 40, 3,  7, 0, 185, 127, 0xf0, 0x00, 32, 2,
			{0, 0, 0, 0, 0, 0}},
[F525][Z37X][DD][DS][ST] = { 40, 4, 15, 0, 194, 255, 0xf0, 0x00, 64, 2,
			{0, 0, 0, 0, 0, 0}},
[F525][Z37X][DD][SS][DT] = { 40, 4, 15, 1, 194, 127, 0xc0, 0x00, 32, 2,
			{0, 0, 0, 0, 0, 0}},
[F525][Z37X][DD][DS][DT] = { 40, 4, 15, 0, 394, 255, 0xf0, 0x00, 64, 2,
			{0, 0, 0, 0, 0, 0}},

[F8][Z47][SD][SS][ST]    = { 26, 3,  7, 0, 242, 63, 0xc0, 0x00, 16, 2,
			{255, 0, 0, 0, 0, 0}},
[F8][Z47][SD][DS][ST]    = { 26, 4, 15, 1, 246,127, 0xc0, 0x00, 32, 2,
			{255, 0, 0, 0, 0, 0}},
[F8][Z47][DD][SS][ST]    = { 52, 4, 15, 0, 242, 127, 0xc0, 0x00, 32, 2,
			{9, 0, 0, 0, 0, 0}},
[F8][Z47][DD][DS][ST]    = { 52, 4, 15, 0, 493, 255, 0xf0, 0x00, 64, 2,
			{9, 0, 0, 0, 0, 0}},
[F8][Z47X][DD][SS][ST]   = { 64, 4, 15, 0, 299, 127, 0xc0, 0x00, 32, 2,
			{4, 0, 0, 0, 0, 0}},
[F8][Z47X][DD][DS][ST]   = { 64, 4, 15, 0, 607, 255, 0xf0, 0x00, 64, 2,
			{4, 0, 0, 0, 0, 0}},

[F525][KAYPRO][DD][SS][ST] = { 40, 3,  7, 0, 194, 63, 0xf0, 0x00, 16, 1,
			{0, 0, 0, 0, 0, 0}},
[F525][KAYPRO][DD][DS][ST] = { 40, 4, 15, 1, 196, 63, 0xc0, 0x00, 16, 1,
			{0, 0, 0, 0, 0, 0}},

[F525][HIDENS][QD][DS][QT] = { 68, 4, 15, 0,1351,1023,0xff, 0xff,256, 2,
			{5, 0, 0, 0, 0, 0}},
};

static void check_heath_label(struct dpb *dpb, struct cpmSuperBlock *ddef) {
		unsigned char lab[128];	//way more than we need (label is 25 bytes at +4)
		int x, cs;
		lseek(ddef->dev.fd, 0L, SEEK_SET);
		read(ddef->dev.fd, lab, sizeof(lab));
		cs = 0;
		for (x = 4; x < 4 + 25; ++x) {
			cs += lab[x];
		}
		if ((cs & 0xff) != 0xff) {
			return;
		}
		// Should be Heath Label...
		int ssz = lab[7] * 128;
#ifndef CPMTOOLS
		fprintf(stderr, "found Heath floppy label\n");
		fprintf(stderr, "Heath DPB   %2d %d %2d %d %4d %3d %02x%02x %2d %d (%d)\n",
			(lab[13] | (lab[14] << 8)),
			lab[15], lab[16], lab[17],
			(lab[18] | (lab[19] << 8)),
			(lab[20] | (lab[21] << 8)),
			lab[22], lab[23],
			(lab[24] | (lab[25] << 8)),
			(lab[26] | (lab[27] << 8)),
			ssz
		);
		fprintf(stderr, "Assumed DPB %2d %d %2d %d %4d %3d %02x%02x %2d %d (%d)\n",
			dpb->spt, dpb->bsh, dpb->bsm, dpb->exm, dpb->dsm, dpb->drm,
			dpb->alv0, dpb->alv1, dpb->cks, dpb->restrk, secSize_m
		);
#endif
		// TODO: more extended format data to check?
		if (ssz == secSize_m && memcmp(dpb, &lab[13], 15) == 0) {
			return;
		}
#ifndef CPMTOOLS
		fprintf(stderr, "overriding assumed DPB\n");
#endif
		secSize_m = ssz;
		memcpy(dpb, &lab[13], 15);
}

static void get_floppy_diskdef(struct cpmSuperBlock *ddef) {
	enum fmts fmt;
	enum medias media = mediaSize_m == 5 ? F525 : F8;
	enum densities dens = SD;
	if (doubleDensity_m > 1) {
		dens = QD;
	} else if (doubleDensity_m > 0) {
		dens = DD;
	}
	enum sides sid = numSides_m > 1 ? DS : SS;
	enum tpis tpi = ST;
	if (media == F525) {
		if (numTracks_m > 80) {
			tpi = QT;
		} else if (numTracks_m > 40) {
			tpi = DT;
		}
	}
	if (media == F8) {
		fmt = Z47;
	} else {
		fmt = Z37;
	}
	if (doubleDensity_m > 0 && secSize_m == 512 && numSectors_m == 9) {
		fmt = MMS;
	} else if (doubleDensity_m > 1  && secSize_m == 512 && numSectors_m == 17) {
		fmt = HIDENS;
	}
	if (doubleDensity_m && secSize_m == 512 &&
			numSectors_m == 10 && interlaced_m == 2) {
		fmt = KAYPRO;
	}
	// can't tell M47 from Z47X... just use Z47X...
	// except by interlaced?
	if (media == F8 && secSize_m == 1024 && numSectors_m == 16) {
		fmt = interlaced_m ? Z47X : M47;
	}
	if (secSize_m == 1024) {
		fmt = media == F8 ? Z47X : Z37X;
	}
	if (mediaSec_m == numSectors_m) {
		fmt = Z17;
	}
	struct dpb *dpb = &floppy_dpbs[media][fmt][dens][sid][tpi];
	if (dpb->pad[0] != 255 && dpb->pad[0] != mediaLat_m) {
		mediaLat_m = 0;
	}
	if (fmt == Z47 || fmt == Z47X || fmt == Z37 || fmt == Z37X) {
		// might override DPB
		check_heath_label(dpb, ddef);
	}
	set_diskdef(dpb, 0, secSize_m, mediaLat_m, ddef);
}

static int get_ydsk_diskdef(struct cpmSuperBlock *ddef) {
	unsigned char buf[128];
	memset(buf, 0, sizeof(buf));
	lseek(ddef->dev.fd, 0L, SEEK_SET);
	size_t rc = read(ddef->dev.fd, buf, sizeof(buf));
	if (rc != sizeof(buf) || strcmp(buf, "<CPM_Disk>") != 0) {
		return -1;
	}

	int x = 32;	// offset of DPB in YAZE disk image header
	// WARNING: only works for little-endian machines
	struct dpb *dpb = (struct dpb *)&buf[x];

	set_diskdef(dpb, 128, 128, 1, ddef);
	return 0;
}

static int get_logdisk_def(struct cpmSuperBlock *ddef, int partno) {
	unsigned char buf[128];
	long end = lseek(ddef->dev.fd, 0L, SEEK_END);
	lseek(ddef->dev.fd, end - sizeof(buf), SEEK_SET);
	size_t rc = read(ddef->dev.fd, buf, sizeof(buf));
	if (rc != sizeof(buf) || !checkHeader(buf, sizeof(buf))) {
		// silently fail if not recognized?
		return -1;
	}
	if (result == 0x7f) {
		// Floppy diskette...
		get_floppy_diskdef(ddef);
	} else {
		// SASI disk, possible partition table.
		lseek(ddef->dev.fd, 0L, SEEK_SET);
		char ptab[512];
		read(ddef->dev.fd, ptab, sizeof(ptab));
		get_sasi_diskdef(ddef, ptab, partno);
	}
	return 0;
}

int autoReadSuper(struct cpmSuperBlock *ddef, const char *format) {
	int partno = 0;
	if (format != NULL) {
		const char *s = strchr(format, ':');
		if (s == NULL) s = format;
		else ++s;
		partno = strtoul(s, NULL, 0);
	}
	memset(((char *)ddef) + sizeof(ddef->dev), 0, sizeof(*ddef) - sizeof(ddef->dev));
	ddef->skew = 1;
	ddef->blksiz = -1;
	ddef->boottrk = -1;
	ddef->secLength = -1;
	ddef->sectrk = -1;
	ddef->tracks = -1;
	ddef->maxdir = -1;
	if (get_ydsk_diskdef(ddef) == 0) return 0;
	if (get_logdisk_def(ddef, partno) == 0) {
		return 0;
	}
	return -1;
}

#ifndef CPMTOOLS
void put_diskdef(struct cpmSuperBlock *ddef) {
	float os = check_os(ddef);
	int dnb = (ddef->maxdir * 32 + ddef->blksiz - 1) / ddef->blksiz;
	int ext = ((ddef->size > 256 ? 8 : 16) * ddef->blksiz) / 16384;
	printf("diskdef\n");
	printf("  seclen %d\n", ddef->secLength);
	printf("  tracks %d\n", ddef->tracks);
	printf("  sectrk %d\n", ddef->sectrk);
	printf("  blocksize %d\n", ddef->blksiz);
	if (ext != ddef->extents) {
		printf("  logicalextents %d\n", ddef->extents);
	}
	printf("  maxdir %d\n", ddef->maxdir);
	// assert(dnb <= dirblks);
	if (dnb < ddef->dirblks) {
		printf("  dirblks %d\n", ddef->dirblks);
	}
	printf("  skew %d\n", ddef->skew);
	printf("  boottrk %d\n", ddef->boottrk);
	printf("  offset %ld\n", ddef->offset);
	printf("  os %s\n", (os < 3.0 ? "2.2" : (os == 3.0 ? "mpm" : "3")));
	printf("end\n");
}

int main(int argc, char **argv) {
	struct cpmSuperBlock diskdef;
	int partno = 0;

	if (argc < 2 || argc > 3) {
		fprintf(stderr, "Syntax: %s <image-file> [<partition-num>]\n", argv[0]);
		exit(1);
	}

	diskdef.dev.fd = open(argv[1], O_RDONLY);
	if (diskdef.dev.fd < 0) {
		perror(argv[1]);
		exit(1);
	}
	if (autoReadSuper(&diskdef, argv[2]) < 0) {
		fprintf(stderr, "Failed to recognize image %s\n", argv[1]);
		close(diskdef.dev.fd);
		exit(1);
	}
	put_diskdef(&diskdef);
	close(diskdef.dev.fd);
	return 0;
}
#endif
