// Copyright (c) 2020 Douglas Miller <durgadas311@gmail.com>
import java.awt.*;
import javax.swing.*;
import java.io.*;
import java.net.*;
import java.lang.reflect.Field;
import java.util.Enumeration;
import java.util.Properties;
import java.util.Arrays;
import java.util.Vector;
import java.lang.reflect.Constructor;
// See: https://fazecast.github.io/jSerialComm/
// Currently using: jSerialComm-2.5.3.jar
import com.fazecast.jSerialComm.*;
// Run with:
//	java -cp <this-jar>:<jSerialComm-jar> CpnetSerialServer [args...]

// Only supports a single client... over a serial port
// Handles both client (Server Dispatch) and server (HostFileBdos/Socket) functions

public class CpnetSerialServer extends ServerDispatch
			implements VirtualUART, Runnable {
	private static CpnetSerialServer thus;
	private static final int vers = 3;

	public static void main(String[] args) {
		thus = new CpnetSerialServer(args);
	}

	Properties props = new Properties();
	int srvId = 0xff;
	SerialPort comm;
	String tty = null;
	int baud = -1;
	String prefix;

	InputStream in;
	OutputStream out;
	int cltId;
	byte[] buf;
	int debug = 0;
	CpnetSerialProtocol proto = null;
	int flowCtrl = SerialPort.FLOW_CONTROL_DISABLED;
	// Support for console feature
	boolean cpnet;	// servicing CP/NET or console?
	boolean oob;	// console in OOB mode?
	OutputStream log;
	Object attObj;
	SerialDevice io;
	boolean io_in;
	boolean io_out;
	java.util.concurrent.LinkedBlockingDeque<Integer> fifo;
	java.util.concurrent.LinkedBlockingDeque<Integer> fifi;
	int modem;
	NetworkBoot nwb;

	public CpnetSerialServer(String[] args) {
		super();
		// TODO: need to redirect stderr...
		String cidstr = null;
		int cid = 0xff;
		String rc = System.getenv("CPNET_CONFIG");
		String log = null;
		if (rc == null) {
			File f = new File("./cpnetrc");
			if (f.exists()) {
				rc = f.getAbsolutePath();
			}
		}
		if (rc == null) {
			rc = System.getProperty("user.home") + "/.cpnetrc";
		}
		String s = props.getProperty("cpnet_log");
		if (s != null) {
			log = s;
		}
		// Command line args override config file...
		for (String arg : args) {
			if (arg.startsWith("conf=")) {
				rc = arg.substring(5);
			}
			if (arg.startsWith("log=")) {
				log = arg.substring(4);
			}
		}
		if (log != null) try {
			FileOutputStream fo = new FileOutputStream(log, false);
			PrintStream ps = new PrintStream(fo, true);
			System.setErr(ps);
		} catch (Exception ee) {}
		System.err.format("CpnetSerialServer v1.%d\n", vers);
		try {
			FileInputStream cfg = new FileInputStream(rc);
			props.load(cfg);
			cfg.close();
			System.err.format("Using config in %s\n", rc);
		} catch(Exception ee) {
			System.err.format("No config file\n");
		}
		s = props.getProperty("cpnet_tty");
		if (s != null) {
			String[] ss = s.split("\\s");
			tty = ss[0];
			if (ss.length > 1) {
				baud = Integer.valueOf(ss[1]);
			}
		}
		s = props.getProperty("cpnet_debug");
		if (s != null) {
			try {
				if (s.length() == 0) {
					debug = 1;
				} else {
					debug = Integer.valueOf(s);
				}
			} catch (Exception ee) {
				System.err.format("cpnet_debug: %s\n", ee.getMessage());
			}
		}
		s = props.getProperty("cpnet_cid");
		if (s != null) {
			cidstr = s;
		}
		String[] pp = null;
		s = props.getProperty("cpnet_protocol");
		if (s != null) {
			pp = s.split("\\s");
		}
		s = props.getProperty("cpnet_flow_control");
		if (s != null) {
			if (s.equalsIgnoreCase("none")) {
				flowCtrl = SerialPort.FLOW_CONTROL_DISABLED;
			} else if (s.equalsIgnoreCase("rts/cts") ||
					s.equalsIgnoreCase("cts/rts")) {
				flowCtrl = SerialPort.FLOW_CONTROL_RTS_ENABLED |
					SerialPort.FLOW_CONTROL_CTS_ENABLED;
			} else if (s.equalsIgnoreCase("xon/xoff") ||
					s.equalsIgnoreCase("xoff/xon")) {
				flowCtrl = SerialPort.FLOW_CONTROL_XONXOFF_IN_ENABLED |
					SerialPort.FLOW_CONTROL_XONXOFF_OUT_ENABLED;
			} else {
				System.err.format("Ignoring invalid flow control \"%s\"\n",
							s);
			}
		}
		for (String arg : args) {
			if (arg.startsWith("tty=")) {
				String[] l = arg.substring(4).split(",");
				tty = l[0];
				if (l.length > 1) {
					baud = Integer.valueOf(l[1]);
				}
			} else if (arg.startsWith("cid=")) {
				cidstr = arg.substring(4);
			} else if (arg.startsWith("proto=")) {
				pp = s.split(",");
			} else if (arg.equalsIgnoreCase("debug")) {
				debug = 1;
			} else if (arg.startsWith("debug=")) {
				try {
					debug = Integer.valueOf(arg.substring(6));
				} catch (Exception ee) {
					System.err.format("cpnet_debug: %s\n", ee.getMessage());
				}
			}
		}
		if (pp != null && pp[0].equalsIgnoreCase("DRI")) {
			proto = new CpnetDRIProtocol(props, debug, pp);
		} else {
			proto = new CpnetSimpleProtocol(props, debug, pp);
		}
		//System.err.format("CP/NET Serial protocol %s\n", proto.getClass().getName());
		if (tty == null) {
			System.err.format("No valid TTY specified or configured\n");
			System.exit(1);
		}
		comm = null;
		if (!tty.equals("stdio")) {
			comm = getPort(tty, baud);
			if (comm == null) {
				System.err.format("TTY %s not found\n", tty);
				System.exit(1);
			}
		}
		try {
			cid = Integer.valueOf(cidstr, 16);
		} catch(Exception ee) {}
		if (cid <= 0 || cid >= 0xff) {
			System.err.format("No valid client ID specified or configured\n");
			System.exit(1);
		}
		cpnet = true;
		s = props.getProperty("cpnet_console");
		if (s != null) {
			if (comm != null) { // real TTY
				// Someone will eventually turn these on?
				comm.clearRTS(); // don't accept data, but
				comm.setDTR();	// indicate we are here...
			}
			attachClass(props, s);
			cpnet = false;	// start in console mode...
			fifo = new java.util.concurrent.LinkedBlockingDeque<Integer>();
			fifi = new java.util.concurrent.LinkedBlockingDeque<Integer>();
			oob = (props.getProperty("cpnet_console_oob") != null);
			s = props.getProperty("cpnet_console_log");
			if (s != null) {
				console_log(s.split("\\s"));
			}
		}
		install_ServerDispatch(props, "cpnet_", cid, 0, 255, null);
		cltId = cid;
		buf = new byte[256 + NetworkServer.mhdrlen];
		if (comm == null) {
			in = System.in;
			out = System.out;
		} else try {
			in = comm.getInputStream();
			out = comm.getOutputStream();
		} catch (Exception ee) {
			ee.printStackTrace();
			System.exit(1);
		}
		if (proto != null) {
			proto.setIO(in, out);
		}
		Thread t = new Thread(this);
		t.start();
	}

	private void console_log(String[] args) {
		boolean append = (args.length > 1 && args[1].equalsIgnoreCase("a"));
		try {
			log = new FileOutputStream(args[0], append);
		} catch (Exception ee) {
			System.err.format("Console log error: %s\n", ee.getMessage());
		}
	}

	private void attachClass(Properties props, String s) {
		String[] args = s.split("\\s");
		Vector<String> argv = new Vector<String>(Arrays.asList(args));
		// try to construct from class...
		try {
			Class<?> clazz = Class.forName(args[0]);
			Constructor<?> ctor = clazz.getConstructor(
					Properties.class,
					argv.getClass(),
					VirtualUART.class);
			// funky "new" avoids "argument type mismatch"...
			attObj = ctor.newInstance(
					props,
					argv,
					(VirtualUART)this);
		} catch (Exception ee) {
			System.err.format("Invalid class in attachment: %s\n", s);
			return;
		}
	}

	private SerialPort getPort(String tty, int baud) {
		try {
			SerialPort serialPort = SerialPort.getCommPort(tty);
			if (serialPort == null) {
				return null;
			}
			// TODO: timeout values...
			if (!serialPort.openPort()) {
				return null;
			}
			if (baud > 0) {
				if (!serialPort.setComPortParameters(baud, 8,
							SerialPort.ONE_STOP_BIT,
							SerialPort.NO_PARITY)) {
					serialPort.closePort();
					return null;
				}
			}
			if (!serialPort.setComPortTimeouts(SerialPort.TIMEOUT_READ_BLOCKING,
					0, 0)) {
				serialPort.closePort();
				return null;
			}
			// TODO: always set this?
			if (flowCtrl != SerialPort.FLOW_CONTROL_DISABLED &&
					!serialPort.setFlowControl(flowCtrl)) {
				System.err.format("Failed to set flow control\n");
			}
			return serialPort;
		} catch (Exception ee) {
			ee.printStackTrace();
			return null;
		}
	}

	private void dumpMsg(char ctx, byte[] buf, int off, int len) {
		String s = "";
		for (int x = 0; x < len; ++x) {
			s += String.format("%02x", buf[off + x] & 0xff);
		}
		System.err.format("%c %s (%d,%d)\n", ctx, s, off, len);
	}

	//////////////////////////////////
	/// VirtualUART implementation ///
	public int available() { return 0; }	// not used?
	public int take() { return 0; }	// not used?
	public boolean ready() {
		return false; // TODO: is tty ready to send?
	}
	public void put(int ch, boolean sleep) {
		// TODO: !sleep?
		//System.err.format("put(%02x)\n", ch);
		if (cpnet) {
			fifi.add(ch);
		} else {
			try {
				out.write(ch);
				out.flush();
			} catch (Exception ee) {
				// TODO: detach?
			}
		}
	}
	public void setModem(int mdm) {
		//System.err.format("setModem(%04x)\n", mdm);
		// TODO: set tty modem controls
		int diff = modem ^ mdm;
		modem = mdm;
		if (comm != null) { // real TTY is active, relay handshake
			// TODO: are these truly reversed?
			if ((diff & VirtualUART.SET_CTS) != 0) {
				// TODO: does our CTS translate to their RTS?
				if ((modem & VirtualUART.SET_CTS) != 0) {
					comm.setRTS();
				} else {
					comm.clearRTS();
				}
			}
			// don't meddle with DSR/DTR?
			if (false && (diff & VirtualUART.SET_DSR) != 0) {
				// TODO: does our DSR translate to their DTR?
				if ((modem & VirtualUART.SET_DSR) != 0) {
					comm.setDTR();
				} else {
					comm.clearDTR();
				}
			}
		}
		// If CTS goes active, flush anything in fifo...
		if ((diff & VirtualUART.SET_CTS) != 0 &&
				(modem & VirtualUART.SET_CTS) != 0) {
			while (fifo.size() > 0) {
				try {
					int c = fifo.take();
					io.write(c);
				} catch (Exception ee) {}
			}
		}
	}
	public int getModem() {
		if (comm == null) {
			return (VirtualUART.GET_RTS | VirtualUART.GET_DTR);
		}
		int mdm = 0;
		// TODO: are these truly reversed?
		if (comm.getDSR()) {
			mdm |= VirtualUART.GET_DTR;
		}
		if (comm.getCTS()) {
			mdm |= VirtualUART.GET_RTS;
		}
		//System.err.format("getModem() %04x\n", mdm);
		return mdm;
	}
	public boolean attach(Object periph) {
		if (attObj != null) {
			return false;
		}
		attObj = periph;
		return true;
	}
	public void detach() {
		// what to do in this case?
	}
	public String getPortId() { return tty; }
	public void attachDevice(SerialDevice io) {
		this.io = io;
		if (io == null) {
			io_in = io_out = false;
			return;
		}
		io_in = ((io.dir() & SerialDevice.DIR_IN) != 0);
		io_out = ((io.dir() & SerialDevice.DIR_OUT) != 0);
		// TODO: is modem control update forthcoming?
	}
	//////////////////////////////////

	private byte[] startBoot(byte[] buf) {
		nwb = NetworkBoot.doStart(props, buf);
		if (nwb == null) {
			return NetworkBoot.error(buf);
		}
		// TODO: some sort of timeout...
		// "nwb = null" when fires...
		return nwb.doNext(buf);
	}

	private void conout(int ch) {
		if (log != null) {
			try {
				log.write(ch);
			} catch (Exception ee) {}
		}
		if ((modem & VirtualUART.SET_CTS) == 0) {
			// TODO: limit fifo size...
			fifo.add(ch);
		} else {
			try {
				io.write(ch);
			} catch (Exception ee) {}
		}
	}

	private void cpnetFalse() {
		cpnet = false;
		// if any conin arrived, pass it along...
		while (fifi.size() > 0) {
			try {
				int c = fifi.take() & 0xff;
				out.write(c);
				out.flush();
			} catch (Exception ee) {
				// TODO: detach?
			}
		}
	}

	private int doCpnet(int ch) {
		int n = 0;
		int buflen = 0;
		cpnet = true;
		if (ch < 0) {
			n = proto.recv(buf, 0, buf.length);
		} else {
			n = proto.recv(ch, buf, 0, buf.length);
		}
		if (n != 0) {
			return n;
		}
		n = (buf[NetworkServer.SIZ] & 0xff) + 1 +
				NetworkServer.mhdrlen;
		if (debug > 0) dumpMsg('>', buf, 0, n);
		byte[] rsp;
		int fmt = buf[NetworkServer.FMT] & 0xff;
		int fnc = buf[NetworkServer.FNC] & 0xff;
		if (fmt == 0xb0) { // NetworkBoot protocol...
			if (nwb == null) {
				rsp = startBoot(buf);
			} else {
				rsp = nwb.doNext(buf);
				if (nwb.done()) {
					nwb = null;
				}
			}
			buflen = (rsp[NetworkServer.SIZ] & 0xff) + 1 +
					NetworkServer.mhdrlen;
		} else if (fmt != 0x00) { // garbage?
			buf[NetworkServer.FMT] |= 1;
			byte src = buf[NetworkServer.SID];
			buf[NetworkServer.SID] = buf[NetworkServer.DID];
			buf[NetworkServer.DID] = src;
			buf[NetworkServer.SIZ] = 2 - 1;
			buf[NetworkServer.DAT] = (byte)0xff;
			buf[NetworkServer.DAT+1] = (byte)12;
			buflen = NetworkServer.mhdrlen + 2;
			rsp = buf;
		} else if (fnc == 255) {
			// Client asking "who am i?"...
			buf[NetworkServer.SID] = (byte)0;
			buf[NetworkServer.DID] = (byte)cltId;
			rsp = buf;
			buflen = NetworkServer.mhdrlen + 1;
			// cpnet = true; // already done, or will be.
		} else if (fnc == 254) {
			// Client notifying "shutdown" (but not exit)
			// TODO: in console mode, revert to !cpnet?
			// TODO: fifi.clear()? or flush to out?
			shutdown();
			if (attObj != null) { // console mode
				cpnetFalse();
			}
			// no response...
			return 0;
		} else if (fnc == 3 || fnc == 4 || fnc == 11) {
			// Networked console support
			rsp = doConsole(fnc, buf, n);
			buflen = (rsp[NetworkServer.SIZ] & 0xff) + 1 +
					NetworkServer.mhdrlen;
		} else {
			// we have a message to process...
			// TODO: verify SID? and do what?
			rsp = sendMsg(buf, n);
			if ((rsp[NetworkServer.FMT] & 0xff) != 0x01) { // must be error...
				buf[NetworkServer.FMT] = 1;
				buf[NetworkServer.SID] = buf[NetworkServer.mdid];
				buf[NetworkServer.DID] = (byte)cltId;
				buf[NetworkServer.SIZ] = (byte)(2 - 1);
				buf[NetworkServer.DAT] = (byte)0xff;
				buf[NetworkServer.DAT + 1] = (byte)0x0c;
				buflen = NetworkServer.mhdrlen + 2;
				rsp = buf;
			} else {
				buflen = (rsp[NetworkServer.SIZ] & 0xff) + 1 +
					NetworkServer.mhdrlen;
			}
		}
		if (debug > 0) dumpMsg('<', rsp, 0, buflen);
		int tries = proto.getRetries();
		do {
			n = proto.send(rsp, NetworkServer.FMT, buflen);
		} while (n > 0 && --tries > 0);
		return n;
	}

	private byte[] doConsole(int fnc, byte[] msg, int len) {
		int ret = 1;
		switch (fnc) {
		case 3:	// raw console input
			// Regardless whether char available...
			// sleep if not.
			// TODO: OK to sleep here?
			try {
				buf[NetworkServer.DAT] = (byte)(fifi.take() & 0xff);
			} catch (Exception ee) {
				buf[NetworkServer.DAT] = (byte)0;
			}
			break;
		case 4:	// raw console output
			conout(buf[NetworkServer.DAT + 1] & 0xff);
			buf[NetworkServer.DAT] = (byte)0;
			break;
		case 11: // get console input status
			buf[NetworkServer.DAT] = (byte)(fifi.size() > 0 ? 255 : 0);
			break;
		default:	// never reached?
			buf[NetworkServer.DAT] = (byte)0xff;
			buf[NetworkServer.DAT + 1] = (byte)0x0c;
			++ret;
			break;
		}
		buf[NetworkServer.FMT] = (byte)1;
		byte sid = buf[NetworkServer.SID];
		buf[NetworkServer.SID] = buf[NetworkServer.DID];
		buf[NetworkServer.DID] = sid;
		buf[NetworkServer.SIZ] = (byte)(ret - 1);
		return buf;
	}

	public void run() {
		while (true) {
			int n = 0;
			// OOB: console output chars may be sent between CP/NET messages
			if (cpnet && !oob) {
				n = doCpnet(-1);
			} else {
				try {
					n = in.read();
					if (proto.protoStart(n)) {
						n = doCpnet(n);
						// TODO: specific errors?
						cpnetFalse();
					} else {
						// can't get here unless io_out?
						conout(n);
					}
				} catch (Exception ee) {
					n = -1;
				}
			}
			if (n < 0) {
				shutdown();
				// TODO: really exit?
				break;
			}
			// should be (n == 0)...
		}
		System.err.format("COMM failure, terminating\n");
		if (comm != null) {
			comm.closePort();
		}
	}
}
