// Copyright (c) 2022 Douglas Miller <durgadas311@gmail.com>

import java.util.Arrays;
import java.util.Properties;
import java.util.Vector;
import java.awt.event.*;
import java.io.*;

public class GIDE_CF implements IODevice, DiskController {

	// implements a range of 16 I/O ports
	static final int adr_CS0_c = 0x08; // 0x00/0x08 indicates where /CS0 lies
	static final int adr_CS1_c = 0x00; // 0x08/0x00 indicates where /CS1 lies
	static final int adr_DrvSel_c = 0; // bits 1,0 select CF1,CF0

	private int base;
	private Interruptor intr;
	private LED[] leds_m;

	private String name;
	private boolean gide;

	private CFCard[] drives;
	private CFCard cd;

	GIDE_CF(Properties props, String lab, int base, LEDHandler lh,
			Interruptor intr) {
		this.intr = intr;
		this.base = base;
		
		drives = new CFCard[2];
		String pfx;
		// Always show drive on front panel, even if not usable
		gide = (lab == null);
		if (gide) {
			// default: GIDE with master/slave
			name = "GIDE_CF";
			pfx = "gide";
		} else {
			// direct attach CFs, separate select
			name = lab.toUpperCase();
			pfx = lab.toLowerCase();
		}
		leds_m = lh.registerLEDs(name, 2, new LED.Colors[] {
				LED.Colors.BLUE, LED.Colors.BLUE});
		String s = props.getProperty(pfx + "_port");
		// TODO: change port
		s = props.getProperty(pfx + "_intr");
		// TODO: change intr
		setupDrive(props, pfx, 0);
		setupDrive(props, pfx, 1);

		if (gide) {
			// default to drive 0 selected...
			selDrv(0);
		}
		System.err.format("%s device at port %02x\n", name, base);
	}

	private void setupDrive(Properties props, String pfx, int ix) {
		long cap = 0;
		String s = props.getProperty(String.format("%s_drive%d", pfx, ix));
		if (s == null && ix == 0) {
			s = props.getProperty("cf_drive");
			if (s == null) {
				s = "CF 64M";
			}
		}
		if (s == null) {
			return;
		}
		String[] ss = s.split("\\s", 3);
		if (ss.length > 1 && ss[1].matches("[0-9]+[MmGg]")) {
			int i = ss[1].length() - 1;
			char f = Character.toUpperCase(ss[1].charAt(i));
			cap = Integer.valueOf(ss[1].substring(0, i)) *
				1024 * 1024;	// Megabytes at least...
			if (f == 'G') {
				cap *= 1024;
			}
		}
		if (ss.length > 2) {
			// Front panel name, not media name
			//name[ix] = ss[2];
		}
		drives[ix] = new CFCard(ix, ss[0], leds_m[ix], cap);
	}

	public GenericDiskDrive findDrive(String name) {
		if (drives[0] != null && name.equals(drives[0].getDriveName())) {
			return drives[0];
		}
		if (drives[1] != null && name.equals(drives[1].getDriveName())) {
			return drives[1];
		}
		return null;
	}
	public Vector<GenericRemovableDrive> getDiskDrives() {
		// Only return REMOVABLE drives...
		Vector<GenericRemovableDrive> drives = new Vector<GenericRemovableDrive>();
		//drives.add(this);
		return drives;
	}

	public int getBaseAddress() { return base; }
	public int getNumPorts() { return 16; }

	public int in(int port) {
		if (cd == null) return 0;
		port &= 0x0f;
		if (port == adr_DrvSel_c) return 0;
		int val = 0;
		if ((port & 0x08) == adr_CS1_c) {
			val = cd.in1(port);
		} else {
			val = cd.in0(port);
		}
		return val;
	}

	public void out(int port, int val) {
		port &= 0x0f;
		if (!gide && port == adr_DrvSel_c) {
			// convert 10b,01b => 1,0; -1,2 for "none"
			selDrv((val & 3) - 1);
			return;
		}
		if (gide && port == adr_CS0_c + CFCard.adr_Head_c) {
			selDrv(CFCard.getLUN((byte)val));
		}
		if (cd == null) return;
		if ((port & 0x08) == adr_CS1_c) {
			cd.out1(port, val);
		} else {
			cd.out0(port, val);
		}
	}

	public void reset() {
		if (drives[0] != null) drives[0].reset();
		if (drives[1] != null) drives[1].reset();
		if (gide) {
			selDrv(0);
		} else {
			selDrv(-1);
		}
	}

	// ix may be -1 or > 1
	private void selDrv(int ix) {
		if (ix < 0 || ix > 1) {
			cd = null;
		} else {
			cd = drives[ix];
		}
		if (cd != null) {
			cd.selDrv();
		}
	}

	public String getDeviceName() { return name; }

	public String dumpDebug() {
		String ret = "";
		if (drives[0] != null) {
			ret += drives[0].dumpDebug();
		}
		if (drives[1] != null) {
			ret += drives[1].dumpDebug();
		}
		return ret;
	}
}
