// Copyright (c) 2016 Douglas Miller <durgadas311@gmail.com>

import java.io.*;
import java.util.Arrays;
import java.util.Properties;

public class GenericSASIDrive implements GenericDiskDrive {
	static final int INVALID = 0;
	// TODO: fix this: controllers vs. drives...
	static final int XEBEC_ST506 = 1;	//  5.3MB (XEBEC ctrlr + Seagate ST506 drive)
	static final int XEBEC_ST412 = 2;	// 10.6MB
	static final int XEBEC_CM5206 = 3;	//  4.4MB
	static final int XEBEC_CM5410 = 4;	//  8.9MB
	static final int XEBEC_CM5616 = 5;	// 13.3MB
	static final int XEBEC_RO201 = 6;	//  5.5MB
	static final int XEBEC_RO202 = 7;	// 11.1MB
	static final int XEBEC_RO203 = 8;	// 16.7MB
	static final int XEBEC_RO204 = 9;	// 22.3MB
	static final int other_other = 10;	// not supported
	static final int NUM_DRV_TYPE = 11;

	public static final int ctl_Busy_o_c = 0x01;
	public static final int ctl_Ack_i_c = 0x02;
	public static final int ctl_Reset_i_c = 0x04;
	public static final int ctl_Msg_o_c = 0x08;
	public static final int ctl_Sel_i_c = 0x10;
	public static final int ctl_Cmd_o_c = 0x20;
	public static final int ctl_Req_o_c = 0x40;
	public static final int ctl_Out_o_c = 0x80;

	// SASI Command codes, class/opcode.
	static final byte cmd_TestDriveReady_c = 0x00;
	static final byte cmd_Recal_c = 0x01;
	static final byte cmd_ReqSense_c = 0x03;
	static final byte cmd_FormatDrive_c = 0x04;
	static final byte cmd_CheckTrackFmt_c = 0x05;
	static final byte cmd_FormatTrack_c = 0x06;
	static final byte cmd_FormatBadTrack_c = 0x07;
	static final byte cmd_Read_c = 0x08;
	static final byte cmd_Write_c = 0x0a;
	static final byte cmd_Seek_c = 0x0b;
	static final byte cmd_InitDriveChar_c = 0x0c;
	static final byte cmd_ReadECCLen_c = 0x0d;
	static final byte cmd_FormatAltTrack_c = 0x0e;
	static final byte cmd_WriteSecBuf_c = 0x0f;
	static final byte cmd_ReadSecBuf_c = 0x10;
	static final byte cmd_RAMDiag_c = (byte)0xe0;
	static final byte cmd_DriveDiag_c = (byte)0xe3;	// was 0xe4 - conflict!
	static final byte cmd_CtrlIntDiag_c = (byte)0xe4;	// conflict!
	static final byte cmd_ReadLong_c = (byte)0xe5;
	static final byte cmd_WriteLong_c = (byte)0xe6;

	private enum State {
		IDLE,
		COMMAND,
		DATA_IN,
		DATA_OUT,
		SENSE,
		DRIVECB,
		STATUS,
	};
	private State curState;

	protected RandomAccessFile[] driveFd;
	protected String[] driveMedia;
	protected int driveSecLen;
	protected int sectorsPerTrack;
	protected long capacity;
	protected int driveCode;
	protected int driveCnum;
	protected long mediaSpt;
	protected long mediaSsz;
	protected long mediaCyl;
	protected long mediaHead;
	protected long mediaLat;
	protected int curLun;

	private int blockCount;

	// mode COMMAND
	protected byte[] cmdBuf = new byte[6];
	private int cmdIx;
	// mode SENSE
	protected byte[] senseBuf = new byte[4];
	private int senseIx;
	// mode DRIVECB
	protected byte[] dcbBuf = new byte[8];
	private int dcbIx;
	// mode STATUS
	protected byte[] stsBuf = new byte[2];
	static final int lastStsIx = 1;
	private int stsIx;
	// mode DATA_IN/DATA_OUT
	private byte[] dataBuf;
	private int dataLength;
	private int dataIx;

	protected String name;

	// Expected drive characteristic by drive: (XEBEC S1410 manual)
	// Drive        cyl  hd red-wr precomp
	// CM5206   256  2  256 256
	// CM5410   256  4  256 256
	// CM5616   256  6  256 256
	// HD561    180  2  128    180
	// HD562    180  2  128    180
	// RMS503   153  2  77  77
	// RMS506   153  4  77  77
	// RMS512   153  8  77  77
	// ST506    153  4  128 64
	// ST412    306  4  128 64
	// TM602S   153  4  128 153
	// TM603S   153  6  128 153
	// TM603SE  230  6  128 128
	// TI5.25+  153  4  64  64
	// RO101    192  2  96  0
	// RO102    192  4  96  0
	// RO103    192  6  96  0
	// RO104    192  8  96  0
	// RO201    321  2  132 0
	// RO202    321  4  132 0
	// RO203    321  6  132 0
	// RO204    321  8  132 0
	// MS1-006  306  2  153 0
	// MS1-012  306  4  153 0
	//
	private static int[][] params = new int[][] {
		{ 0, 0, 0, 0 },		// /* INVALID=0    */
		{ 153, 4, 128, 64 },	// /*[XEBEC_ST506] */
		{ 306, 4, 128, 64 },	// /*[XEBEC_ST412] */
		{ 256, 2, 256, 256 },	// /*[XEBEC_CM5206]*/
		{ 256, 4, 256, 256 },	// /*[XEBEC_CM5410]*/
		{ 256, 6, 256, 256 },	// /*[XEBEC_CM5616]*/
		{ 321, 2, 132, 0 },	// /*[XEBEC_RO201] */
		{ 321, 4, 132, 0 },	// /*[XEBEC_RO202] */
		{ 321, 6, 132, 0 },	// /*[XEBEC_RO203] */
		{ 321, 8, 132, 0 },	// /*[XEBEC_RO204] */
	};

	protected int[] devCfg;

	// All LUNs must be identical
	GenericSASIDrive(Properties props, int type, String name,
			String ctrl, String dir, int cnum, int sectorSize) {
		this.name = name;
		curState = State.IDLE;
		driveFd = null;
		driveSecLen = 0;
		sectorsPerTrack = 0;
		capacity = 0;
		driveCode = 0;
		driveCnum = 0;
		driveMedia = null;
		mediaSpt = 0;
		mediaSsz = 0;
		mediaCyl = 0;
		mediaHead = 0;
		mediaLat = 0;
		cmdIx = 0;
		senseIx = 0;
		dcbIx = 0;
		stsIx = 0;
		dataBuf = null;
		dataLength = 0;
		dataIx = 0;

		devCfg = params[type];
		driveCnum = cnum;
		driveSecLen = sectorSize;

		// These get changed if media exists...
		if (driveSecLen == 256) {
			sectorsPerTrack = 32;
		} else if (driveSecLen == 512) {
			sectorsPerTrack = 17;
		} else {
			// not supported, but do something...
			if (driveSecLen == 0) {
				sectorsPerTrack = 0;
			} else {
				sectorsPerTrack = 8192 / driveSecLen;
			}
		}

		// loop through properties looking for tolower(ctrl) + "_disk%d.%d"...
		String prefix = ctrl.toLowerCase();
		if (props != null) {
			// Maximum of four LUNs...
			// TODO: make sure all are identical...
			int num = 0;
			driveFd = new RandomAccessFile[4];
			driveMedia = new String[4];
			// Legacy property name for LUN 0...
			int x = 0;
			String p = String.format("%s_disk%d", prefix, cnum);
			p = props.getProperty(p);
			if (p != null) {
				driveMedia[x] = p;
				driveFd[x] = getLUN(p, dir, x);
				++num;
				++x;
			}
			for (; x < 4; ++x) {
				p = String.format("%s_disk%d.%d", prefix, cnum, x);
				p = props.getProperty(p);
				if (p == null) {
					continue;
				}
				++num;
				driveMedia[x] = p;
				driveFd[x] = getLUN(p, dir, x);
			}
			if (num == 0) {
				// Legacy behavior...
				p = String.format("%s-%d", ctrl, cnum);
				driveFd[0] = getLUN(p, dir, 0);
				driveMedia[0] = p;
			}
		}
		dataBuf = new byte[driveSecLen + 4];	// space for ECC for "long" commands
		dataLength = driveSecLen;
	}

	protected RandomAccessFile getLUN(String media, String dir, int lun) {
		RandomAccessFile fd;
		if (media.charAt(0) != '/') {
			media = String.format("%s/%s", dir, media);
		}
		try {
			fd = new RandomAccessFile(media, "rw");
			byte[] buf = new byte[128];
			if (fd.length() == 0) {
				capacity = (long)devCfg[0] * devCfg[1] *
							sectorsPerTrack * driveSecLen;
				// special case: new media - initialize it.
				mediaCyl = devCfg[0];
				mediaHead = devCfg[1];
				mediaSsz = driveSecLen;
				mediaSpt = sectorsPerTrack;
				mediaLat = 1;
				Arrays.fill(buf, (byte)0);
				String hdr = String.format("%dc%dh%dz%dp%dl\n",
					mediaCyl, mediaHead, mediaSsz, mediaSpt, mediaLat);
				memcpy(buf, hdr.getBytes(), 0, buf.length);
				// NOTE: 'buf' includes a '\n'...
				System.err.format("Initializing new media %s CTL %d LUN %d as %s",
					media, driveCnum, lun, hdr);
				fd.setLength(capacity + buf.length);
				fd.seek(capacity);	// i.e. END - sizeof(buf)
				fd.write(buf);
			} else {
				fd.seek(fd.length() - buf.length);
				int n = fd.read(buf);
				if (n != buf.length || !checkHeader(buf)) {
					System.err.format("GenericSASIDrive: file %s is not GenericSASIDrive\n", media);
					return null;
				}
				// override sector size from media...
				driveSecLen = (int)mediaSsz;
				sectorsPerTrack = (int)mediaSpt;
				if (devCfg[0] != mediaCyl ||
				    devCfg[1] != mediaHead) {
					// TODO: fatal or warning?
					System.err.format("Media/Drive mismatch: %s\n", media);
					return null;
				}
				capacity = (long)mediaCyl * mediaHead * mediaSpt * mediaSsz;
				String s = new String(buf);
				int i = s.indexOf('\000');
				if (i < 0) i = s.length();
				System.err.format("Mounted existing media %s CTL %d LUN %d as %s",
					media, driveCnum, lun, s.substring(0, i));
			}
			return fd;
		} catch (Exception ee) {
			System.err.format("GenericSASIDrive: Unable to open media %s\n", media);
			return null;
		}
	}

	static public GenericSASIDrive getInstance(Properties props,
			String type, String name, String ctrl, String dir, int cnum) {
		int etype;
		int ssz = 512;

		if (type.equals("XEBEC_ST506")) {
			etype = XEBEC_ST506;
		} else if (type.equals("XEBEC_ST412")) {
			etype = XEBEC_ST412;
		} else if (type.equals("XEBEC_CM5206")) {
			etype = XEBEC_CM5206;
		} else if (type.equals("XEBEC_CM5410")) {
			etype = XEBEC_CM5410;
		} else if (type.equals("XEBEC_CM5616")) {
			etype = XEBEC_CM5616;
		} else if (type.equals("XEBEC_RO201")) {
			etype = XEBEC_RO201;
		} else if (type.equals("XEBEC_RO202")) {
			etype = XEBEC_RO202;
		} else if (type.equals("XEBEC_RO203")) {
			etype = XEBEC_RO203;
		} else if (type.equals("XEBEC_RO204")) {
			etype = XEBEC_RO204;
		} else if (type.startsWith("Z67IDE_")) {
			// get CF capacity, assume 2 cards.
			int e = type.length();
			if (type.charAt(e - 1) == 'G') {
				--e;
			}
			int cap = Integer.valueOf(type.substring(7, e));
			if (cap < 1 || cap > 16 || (cap & (cap - 1)) != 0) {
				// invalid CF card capacity...
			}
			return new Z67IDESASIDrive(props, name, ctrl, dir, cnum, cap);
		} else {
			System.err.format("Invalid controller/drive combination: %s\n", type);
			return null;
		}

		/*
		 * port determination based on SW501...
		    324         MVI     B,7CH
		    325         IN      GPIO            ; READ SWITCH 501
		    326         ANI     00000011B       ; WHAT'S PORT 7C SET FOR ?
		    327         CPI     00000010B       ;  IF Z67, THEN THIS IS IT
		    328         JRZ     GOTPRT
		    329         MVI     B,78H
		    330         IN      GPIO            ; READ SWITCH 501
		    331         ANI     00001100B       ; WHAT'S PORT 78 SET FOR ?
		    332         CPI     00001000B       ;  IF Z67, THEN THIS IS IT
		    333         RNZ
		 */

	//         DB      0,153,4,0,128,0,64,11   ; DRIVE CHARACTERISTIC DATA
	// =   0099, 04, 0080, 0064, 0b
	// =   153 cyl, 4 heads, 128 red-wr, 100 wr-pre, 11 ecc-burst
	// =   612 trk. (~16 spt, 512B ea, ?)
		// ssz = jumper_W2 ? 512 : 256; // XEBEC jumper "W2" a.k.a. "SS" pos "2" or "5"
		// Use sector size specified in image... if existing
		return new GenericSASIDrive(props, etype, name, ctrl, dir, cnum, ssz);
	}

	private void memcpy(byte[] dst, byte[] src, int start, int len) {
		int end = start + len;
		for (int x = start; x < src.length && x < dst.length && x < end; ++x) {
			dst[x] = src[x];
		}
	}

	private boolean checkHeader(byte[] buf) {
		int m = 0;
		int e = 0;
		while (buf[e] != '\n' && buf[e] != '\0' && e < buf.length) {
			int p = 0;
			while (e < buf.length && Character.isDigit(buf[e])) {
				p = (p * 10) + (buf[e] - '0');
				++e;
			}
			if (e >= buf.length) {
				// too harsh?
				return false;
			}
			// TODO: removable flag, others?
			// NOTE: removable media requires many more changes.
			switch (Character.toLowerCase(buf[e])) {
			case 'c':
				if (p > 0x0ffff) {
					break;
				}
				m |= 0x01;
				mediaCyl = p & 0x0ffff;
				break;

			case 'h':
				if (p > 0x0ff) {
					break;
				}
				m |= 0x02;
				mediaHead = p & 0x0ff;
				break;

			case 'z':
				if (p != 128 && p != 256 && p !=512 && p != 1024) {
					break;
				}
				m |= 0x04;
				mediaSsz = p & 0x0fff;
				break;

			case 'p':
				if (p > 0x0ff) {
					break;
				}
				m |= 0x08;
				mediaSpt = p & 0x0ff;
				break;

			case 'l':
				if (p > 0x0ff) {
					break;
				}
				m |= 0x10;
				mediaLat = p & 0x0ff;
				break;

			default:
				return false;
			}

			e++;
		}
		return m == 0x1f;
	}

//
// Typical sequence:
//
// "Hail" controller:
//              wait !BSY (while poking data reg)
//              SEL
//              wait BSY
//              RUN
//
// Send command:
//              wait until REQ+CMD+POUT+BSY
//              out next cmd byte
//              loop
//
// Read command                        Write command
//             if loose POUT, goto status
//              wait for !CMD (loop)
// inir 128 bytes                      outir 128 bytes
//              loop
//
// Status phase:
//               check MSG+REQ+CMD+POUT
//               if REQ+CMD then input data and loop
//               if not MSG+REQ+CMD loop (no data read)
//               input (discard) data
//               return previous data byte (has error status)
//
// MSG    REQ    CMD    POUT    BSY
// ---------------------------------
// X      X      X      X       -       OK to begin
//                                      SEL
// X      X      X      X      BSY      controller ready
//                                      RUN
//
// X     REQ    CMD    POUT    BSY      OK to send command
//                                      OUT cmd byte (ACK)
//                                      loop
//
// X     REQ    CMD     -      BSY      done with data (status begins)
// X     REQ     -      X      BSY      OK to send/get data (bulk 128 bytes)
//                                      IN/OUT data byte (ACK)
//
// -     REQ    CMD     -       X       OK get status byte
//                                      IN status byte (ACK)
//
// MSG    REQ    CMD     -       X      OK last status byte (prev has error)
//                                      IN status byte (ACK)
// presumably resturns to:
// X     REQ    CMD    POUT    BSY      OK to send command.
//
//

	void deselect(SASIDriveController bus) {
		// User is switch controllers/drive, do any cleanup.
		bus.ctrlBus_m &= ~ctl_Msg_o_c;
		bus.ctrlBus_m &= ~ctl_Cmd_o_c;
		bus.ctrlBus_m &= ~ctl_Out_o_c;
		bus.ctrlBus_m &= ~ctl_Req_o_c;
		bus.ctrlBus_m &= ~ctl_Busy_o_c;
		curState = State.IDLE;
	}

	void ack(SASIDriveController bus) {
		bus.ctrlBus_m &= ~ctl_Req_o_c;
		bus.ctrlBus_m &= ~ctl_Ack_i_c;

		if (curState == State.COMMAND) {
			cmdBuf[cmdIx++] = (byte)bus.dataOutReg_m;
			if (cmdIx >= cmdBuf.length) {
				curState = State.IDLE;// default? should get set to something else.
				bus.ctrlBus_m &= ~ctl_Cmd_o_c;
				cmdIx = 0;
				blockCount = 0;
				curLun = (cmdBuf[1] >> 5) & 0x07;
				// this further modifies 'ctrl' and sets curState.
				processCmd(bus);
				return;
			}
			bus.ctrlBus_m |= ctl_Req_o_c;
			bus.ctrlBus_m &= ~ctl_Ack_i_c;
		} else if (curState == State.STATUS) {
			if (stsIx < stsBuf.length) {
				bus.dataInReg_m = stsBuf[stsIx++] & 0xff;
				if (stsIx >= stsBuf.length) {
					bus.ctrlBus_m |= ctl_Msg_o_c;
				}
				bus.ctrlBus_m |= ctl_Req_o_c;
				bus.ctrlBus_m &= ~ctl_Ack_i_c;
				return;
			}
			// must start over at SEL again...
			bus.ctrlBus_m &= ~ctl_Msg_o_c;
			bus.ctrlBus_m &= ~ctl_Busy_o_c;// it sppears this must go low...
			bus.ctrlBus_m &= ~ctl_Cmd_o_c;
			bus.ctrlBus_m &= ~ctl_Out_o_c;
			bus.ctrlBus_m &= ~ctl_Req_o_c;
			curState = State.IDLE;
		} else if (curState == State.SENSE) {
			if (senseIx < senseBuf.length) {
				bus.dataInReg_m = senseBuf[senseIx++] & 0xff;
				bus.ctrlBus_m |= ctl_Req_o_c;
				bus.ctrlBus_m &= ~ctl_Ack_i_c;
				return;
			}
			senseIx = 0;
			startStatus(bus);
			ack(bus);
		} else if (curState == State.DATA_IN) {
			if (dataIx < dataLength) {
				bus.dataInReg_m = dataBuf[dataIx++] & 0xff;
				bus.ctrlBus_m |= ctl_Req_o_c;
				bus.ctrlBus_m &= ~ctl_Ack_i_c;
				return;
			}
			// We must either put some data in the reg or lower BSY...
			dataIx = 0;
			if (++blockCount < (cmdBuf[4] & 0xff)) {
				// curLun already set...
				processCmd(bus);
				return;
			}
			startStatus(bus);
			ack(bus);
		} else if (curState == State.DRIVECB) {
			dcbBuf[dcbIx++] = (byte)bus.dataOutReg_m;
			if (dcbIx >= dcbBuf.length) {
				dcbIx = 0;
				processDCB(bus);
				return;
			}
			bus.ctrlBus_m |= ctl_Req_o_c;
			bus.ctrlBus_m &= ~ctl_Ack_i_c;
		} else if (curState == State.DATA_OUT) {
			dataBuf[dataIx++] = (byte)bus.dataOutReg_m;
			if (dataIx >= dataLength) {
				dataIx = 0;
				processData(bus);
				return;
			}
			bus.ctrlBus_m |= ctl_Req_o_c;
			bus.ctrlBus_m &= ~ctl_Ack_i_c;
		} else {
			// must be IDLE, and host is trying to sync-up with us.
			// make everything look like we are not here...
			bus.ctrlBus_m &= ~ctl_Msg_o_c;
			bus.ctrlBus_m &= ~ctl_Cmd_o_c;
			bus.ctrlBus_m &= ~ctl_Out_o_c;
			bus.ctrlBus_m &= ~ctl_Req_o_c;
			bus.ctrlBus_m &= ~ctl_Busy_o_c;
		}
	}

	void seek(long off) throws Exception {
		driveFd[curLun].seek(off);
	}

	void resetSASI(SASIDriveController bus) {
		bus.ctrlBus_m &= ~ctl_Msg_o_c;
		bus.ctrlBus_m &= ~ctl_Cmd_o_c;
		bus.ctrlBus_m &= ~ctl_Out_o_c;
		bus.ctrlBus_m &= ~ctl_Req_o_c;
		bus.ctrlBus_m &= ~ctl_Busy_o_c;
		curState = State.IDLE;
	}

	void select(SASIDriveController bus) {
		// validate data bit with cnum...
		int i = Integer.numberOfTrailingZeros(bus.dataOutReg_m);
		if (i > 7) {
			return;
		}
		if (i == driveCnum) {
			bus.ctrlBus_m |= ctl_Busy_o_c;
		}
	}

	void run(SASIDriveController bus) {
		bus.ctrlBus_m |= ctl_Busy_o_c;
		bus.ctrlBus_m |= ctl_Cmd_o_c;
		bus.ctrlBus_m |= ctl_Out_o_c;
		bus.ctrlBus_m |= ctl_Req_o_c;
		cmdIx = 0;
		curState = State.COMMAND;
	}

	void startStatus(SASIDriveController bus) {
		stsIx = 0;
		curState = State.STATUS;
		bus.ctrlBus_m &= ~ctl_Msg_o_c;
		bus.ctrlBus_m |= ctl_Cmd_o_c;
		bus.ctrlBus_m &= ~ctl_Out_o_c;
		bus.ctrlBus_m |= ctl_Req_o_c;
	}

	void startSense(SASIDriveController bus) {
		senseIx = 0;
		curState = State.SENSE;
		bus.ctrlBus_m &= ~ctl_Msg_o_c;
		bus.ctrlBus_m |= ctl_Cmd_o_c;
		bus.ctrlBus_m &= ~ctl_Out_o_c;
		bus.ctrlBus_m |= ctl_Req_o_c;
	}

	void startDataIn(SASIDriveController bus) {
		dataIx = 0;
		curState = State.DATA_IN;
		bus.ctrlBus_m &= ~ctl_Msg_o_c;
		bus.ctrlBus_m &= ~ctl_Cmd_o_c;
		bus.ctrlBus_m &= ~ctl_Out_o_c;
		bus.ctrlBus_m |= ctl_Req_o_c;
	}

	void startDataOut(SASIDriveController bus) {
		dataIx = 0;
		curState = State.DATA_OUT;
		bus.ctrlBus_m &= ~ctl_Msg_o_c;
		bus.ctrlBus_m &= ~ctl_Cmd_o_c;
		bus.ctrlBus_m |= ctl_Out_o_c;
		bus.ctrlBus_m |= ctl_Req_o_c;
	}

	void startError(SASIDriveController bus, int err) {
		stsBuf[0] = 0b00000010;
		stsBuf[1] = 0;
		senseBuf[0] = (byte)err;
		startStatus(bus);
	}

	void startDCB(SASIDriveController bus) {
		dcbIx = 0;
		curState = State.DRIVECB;
		bus.ctrlBus_m &= ~ctl_Msg_o_c;
		bus.ctrlBus_m &= ~ctl_Cmd_o_c;
		bus.ctrlBus_m |= ctl_Out_o_c;
		bus.ctrlBus_m |= ctl_Req_o_c;
	}

	void processCmd(SASIDriveController bus) {
		long off;
		long e;

		if (curLun >= driveFd.length || driveFd[curLun] == null) {
			startError(bus, 0x84);	// drive not ready + addr valid
			ack(bus);
			return;
		}
		if (cmdBuf[0] != cmd_ReqSense_c) {
			Arrays.fill(senseBuf, (byte)0);
		}
		switch (cmdBuf[0]) {
		case cmd_TestDriveReady_c:
		case cmd_Recal_c:
		case cmd_RAMDiag_c:
		case cmd_DriveDiag_c:
		case cmd_CtrlIntDiag_c:
			// no-op: just return success
			stsBuf[0] = 0;
			stsBuf[1] = 0;
			stsIx = 0;
			startStatus(bus);
			ack(bus);
			break;
		case cmd_ReqSense_c:
			// sense data was set by previous command...
			senseIx = 0;
			startSense(bus);
			ack(bus);
			break;
		case cmd_Read_c:
		case cmd_ReadLong_c:
			memcpy(senseBuf, cmdBuf, 1, 3);
			off = ((((((cmdBuf[1] & 0x1f) << 8) | (cmdBuf[2] & 0xff)) << 8) | (cmdBuf[3] & 0xff)) +
			       blockCount) * driveSecLen;
			if (off >= capacity) {
				startError(bus, 0x21);	// illegal disk address
				ack(bus);
				break;
			}
			try {
				seek(off);
				// dataBuf includes (fake) ECC, must limit read()...
				e = driveFd[curLun].read(dataBuf, 0, dataLength);
				if (e != dataLength) {
					startError(bus, 0x94);	// target sector not found + addr valid
					ack(bus);
					break;
				}
			} catch (Exception ee) {
				startError(bus, 0x94);	// target sector not found + addr valid
				ack(bus);
				break;
			}
			dataLength = driveSecLen;
			if (cmdBuf[0] == cmd_ReadLong_c) {
				dataLength += 4;
				// TODO: must we compute ECC?
				dataBuf[dataIx++] = 0;
				dataBuf[dataIx++] = 0;
				dataBuf[dataIx++] = 0;
				dataBuf[dataIx++] = 0;
			}
			startDataIn(bus);
			dataIx = 0;
			// a little risky to recurse here, but startDataIn() should
			// ensure we can't loop.
			ack(bus);
			break;
		case cmd_Write_c:
		case cmd_WriteLong_c:
			memcpy(senseBuf, cmdBuf, 1, 3);
			dataLength = driveSecLen;
			if (cmdBuf[0] == cmd_WriteLong_c) {
				dataLength += 4;
			}
			dataIx = 0;
			// could validate address...
			startDataOut(bus);
			break;
		case cmd_WriteSecBuf_c:
			dataIx = 0;
			dataLength = driveSecLen;
			startDataOut(bus);
			break;
		case cmd_ReadSecBuf_c:
			dataIx = 0;
			dataLength = driveSecLen;
			startDataIn(bus);
			ack(bus);
			break;
		case cmd_InitDriveChar_c:
			startDCB(bus);
			break;
		case cmd_FormatDrive_c:
		case cmd_FormatTrack_c:
		case cmd_FormatBadTrack_c:
		case cmd_CheckTrackFmt_c:
		case cmd_Seek_c:
		case cmd_FormatAltTrack_c:
			// validate address, but otherwise just return success.
			blockCount = 0;
			off = ((((((cmdBuf[1] & 0x1f) << 8) | (cmdBuf[2] & 0xff)) << 8) | (cmdBuf[3] & 0xff)) +
				      blockCount) * driveSecLen;

			if (off >= capacity) {
				startError(bus, 0x21);	// illegal disk address
				ack(bus);
				break;
			}
			stsBuf[0] = 0;
			stsBuf[1] = 0;
			stsIx = 0;
			startStatus(bus);
			ack(bus);
			break;
		case cmd_ReadECCLen_c:
			// TODO: transfer 1 data byte, but this command is only valid
			// after a correctable ECC data error 0x18, which we never return.
			stsBuf[0] = 0;
			stsBuf[1] = 0;
			stsIx = 0;
			startStatus(bus);
			ack(bus);
			break;
		default:
			break;
		}

	}

	void processData(SASIDriveController bus) {
		long off;
		long e;

		switch (cmdBuf[0]) {
		case cmd_Write_c:
			off = ((((((cmdBuf[1] & 0x1f) << 8) | (cmdBuf[2] & 0xff)) << 8) | (cmdBuf[3] & 0xff)) +
			       blockCount) * driveSecLen;
			if (off >= capacity) {
				startError(bus, 0x21);	// illegal disk address
				ack(bus);
				break;
			}
			try {
				seek(off);
				// dataBuf includes (fake) ECC, must limit write()...
				driveFd[curLun].write(dataBuf, 0, dataLength);
			} catch (Exception ee) {
				startError(bus, 0x94);	// target sector not found + addr valid
				ack(bus);
				break;
			}
			if (++blockCount >= (cmdBuf[4] & 0xff)) {
				startStatus(bus);
				ack(bus);
				break;
			}
			startDataOut(bus);
			break;
		case cmd_WriteSecBuf_c:
			// just leave sector buf with data for subsequent command(s)
			startStatus(bus);
			ack(bus);
			break;
		case cmd_WriteLong_c:
			break;
		default:
			break;
		}
	}

	void processDCB(SASIDriveController bus) {
		switch (cmdBuf[0]) {
		case cmd_InitDriveChar_c: {
			// validate? update params?
			int cyl = ((dcbBuf[0] & 0xff) << 8) | (dcbBuf[1] & 0xff);
			int hds = dcbBuf[2] & 0xff;
			int rwc = ((dcbBuf[3] & 0xff) << 8) | (dcbBuf[4] & 0xff);
			int wpc = ((dcbBuf[5] & 0xff) << 8) | (dcbBuf[6] & 0xff);
			// Don't care about ECC length = dcbBuf[7] & 0x0f;

			if (cyl != devCfg[0] || hds != devCfg[1] ||
			    rwc != devCfg[2] || wpc != devCfg[3]) {
				System.err.format("Host drive characteristics " +
					" mismatch\n");
				// TODO: what to do. Host will be sending commands based
				// on a different geometry. Can we return an error?
				// startError(bus, 0x??);
			}
			startStatus(bus);
			ack(bus);
			} break;
		default:
			break;
		}
	}

	public String getMediaName() {
		return ""; // Not used, yet.
	}

	public boolean isReady() {
		return true; // Not used, yet.
	}

	public void insertMedia(String[] media) { }
	public int getRawBytesPerTrack() { return 0; }
	public int getNumTracks() { return 0; }
	public int getNumHeads() { return 0; } 
	public int getMediaSize() { return 0; }
	public boolean isRemovable() { return false; }

	public String getDriveName() {
		return name;
	}

	public String dump() {
		String ret = "";
		ret += String.format("Ctrl %d: SPT %d, SSZ %d\n", driveCnum, sectorsPerTrack, driveSecLen);
		ret += String.format("        DEV: %04x %02x %04x %04x\n",
			devCfg[0], devCfg[1], devCfg[2], devCfg[3]);
		ret += String.format("        CMD: %02x %02x %02x %02x %02x %02x\n",
			cmdBuf[0], cmdBuf[1], cmdBuf[2], cmdBuf[3], cmdBuf[4], cmdBuf[5]);
		ret += String.format("        State: %s dataIx=%d blockCount=%d\n",
				curState.name(), dataIx, blockCount);
		return ret;
	}
}
