// Copyright (c) 2018 Douglas Miller <durgadas311@gmail.com>

import java.util.Vector;
import java.util.Properties;

// To simplify internal interfaces, treat as single set of bits
// no matter how many ports.
//
// H8:
//      |  7  |  6  |  5  |  4  |  3  |  2  |  1  |  0  |
//      |-----|-----|-----|-----|-----|-----|-----|-----|
// F0H  |beep |int1 | mon |int2 |  D  |  C  |  B  |  A  | (LSbyte)
//      |-----|-----|-----|-----|-----|-----|-----|-----|
// F1H  | dp  |  f  |  e  |  d  |  c  |  b  |  a  |  g  |
//      |-----|-----|-----|-----|-----|-----|-----|-----|
// F2H* |     | ss  |org0 |     |     |     |  *  |     | (MSbyte)
//      |-----|-----|-----|-----|-----|-----|-----|-----|
//
// * = 2mS clock on Norberto's Z80 CPU board


public class H8Gpp implements H8Fpp {
	// keyboard switch states are maintained in low 16 bits of 'gpi'
	private int gpi;
	private int gpo;
	private boolean change;
	private Interruptor intr;
	private int src;
	private Vector<GppListener> notif;
	private FPListener fpl;
	static final int org0Bit_c = (1 << 21);
	static final int mSClockBit = (1 << 6);
	static final int sStepBit = (1 << 4);
	static final int sideSelBit_c = (1 << 22);

	public H8Gpp(Properties props, Interruptor intr) {
		this.intr = intr;
		src = intr.registerINT(1);
		gpi = 0;
		notif = new Vector<GppListener>();
		String s = props.getProperty("sw1");
		if (s != null) {
			gpi |= (Integer.valueOf(s, 2) & 0xff) << 16;
		}
		reset();
	}

	// For GUI, Keyboard switches
	public void setKey(int kn) {
		kn &= 0x0000ff;
		kn ^= 0x0000ff;
		gpi &= ~0x0000ff;
		gpi |= kn;
//System.err.format("keypad %06x\n", gpi);
	}

	public int getOrg0Bit() { return org0Bit_c; }
	public int get2mSClockBit() { return mSClockBit; }
	public int getSStepBit() { return sStepBit; }
	public boolean isSStepEna() { return (gpo & sStepBit) == 0; }
	public int getSideSelBit() { return sideSelBit_c; }
	public int getMEM1Bit() { return (1 << 19); } // does not exist?
	public int getIO1Bit() { return (1 << 23); } // does not exist?

	/// IODevice interface ///

	public void reset() {
		gpi |= 0x0000ff;
		gpo = 0x0000df;	// databus inverted, tricky
		change = false;
		notify(0xffffff); // might contradict device resets?
	}

	public int getBaseAddress() {
		return 0xf0;
	}
	public int getNumPorts() {
		return 3;
	}
	public String getDeviceName() { return null; }

	public int in(int port) {
		int value = gpi;
		int off = port - getBaseAddress();
		value >>= (off * 8);
		value &= 0xff;
		return value;
	}
	public void out(int port, int value) {
//System.err.format("H8 gpio out %02x => %02x\n", value, port);
		int off = port - getBaseAddress();

		if (off == 2) {
			value &= 0x30;
		} else {
			// TODO: only mark 'change' if display bits change
			change = true;
			value &= 0xff;
		}
		int mask = (0xff << (off * 8));
		value <<= (off * 8);
		
		if ((mSClockBit & mask) != 0) {
			intr.lowerINT(1, src);
		}
		int nv = gpo;
		nv &= ~mask;
		nv |= value;
		newValue(nv);
	}

	private void newValue(int nv) {
		int diff = gpo ^ nv;
		gpo = nv;
		notify(diff);	// Assume Front Panel is a listener...
	}

	private void notify(int diff) {
		if (diff == 0) {
			return;
		}
		for (GppListener lstn : notif) {
			if ((diff & lstn.interestedBits()) != 0) {
				lstn.gppNewValue(gpo);
			}
		}
	}

	public void addGppListener(GppListener lstn) {
		notif.add(lstn);
		lstn.gppNewValue(gpo);
	}

	public void setFPListener(FPListener lstn) {
		fpl = lstn;
	}

	public String dumpDebug() {
		String ret = new String();
		ret += String.format("GP IN: %02x %02x %02x\n",
					(gpi >> 16) & 0xff,
					(gpi >> 8) & 0xff,
					gpi & 0xff);
		ret += String.format("GP OUT: %02x %02x %02x\n",
					(gpo >> 16) & 0xff,
					(gpo >> 8) & 0xff,
					gpo & 0xff);
		ret += String.format("2mS Clock: %s\n", ((gpo & mSClockBit) != 0));
		ret += String.format("ORG0: %s\n", ((gpo & org0Bit_c) != 0));
		return ret;
	}

	public void trigger2ms() {
		if (fpl != null) {
			// actually, at t-0.5 mS...
			fpl.update();
			if (change) {
				change = false;
				fpl.timeout(gpo);
			}
		}
		newValue((gpo & 0xffffdf) | 0x00ff00);
		if ((gpo & mSClockBit) != 0) {
			intr.raiseINT(1, src);
		}
	}
}
