// Copyright (c) 2017 Douglas Miller <durgadas311@gmail.com>

import java.util.Properties;
import java.util.Arrays;
import java.io.*;

public class H8_512KMemory extends H89Roms implements IODevice, Memory, GppListener {
	int base;
	boolean rom;
	boolean map;
	int gpioBit;
	int gpioORG0;
	GeneralPurposePort gpp;
	// Organized as [32][16384]...
	private byte[] mem;
	private int rdBank[];
	private int wrBank[];

	public H8_512KMemory(Properties props, int port,
			GeneralPurposePort gpio, Interruptor intr) {
		super(props, gpio, intr);
		gpp = gpio;
		base = port;
		String s = props.getProperty("h8_512k_port");
		if (s != null) {
			base = Integer.decode(s) & 0xff;
		}
		mem = new byte[32*16384]; // 32x16K = 512K
		rdBank = new int[4]; // 4 - 16K pages in 64K addr space
		wrBank = new int[4]; // 4 - 16K pages in 64K addr space
		reset();
		gpioORG0 = gpio.getOrg0Bit();
		gpioBit = romBit | gpioORG0;
		gpp.addGppListener(this);
		System.err.format("H8-512K Memory Controller at port %02x\n", base);
	}

	public int getBaseAddress() { return base; }
	public int getNumPorts() { return 8; }
	public int in(int port) {
		return 0xff;
	}
	public void out(int port, int value) {
//System.err.format("H8-512K OUT %02x: %02x\n", port, value);
		int x = port & 0x03;
		map = ((value & 0x80) != 0);
		value &= 0x1f;
		boolean wr = ((port & 0x04) != 0);
		if (wr) {
			wrBank[x] = value << 14;
		} else {
			rdBank[x] = value << 14;
		}
	}
	public String getDeviceName() { return "H8-512K"; }
	public String dumpDebug() {
		String ret = String.format("ROM: %s   MAP: %s\n", rom, map);
		ret +=               "            --RD-   --WR-\n";
		ret += String.format("PAGE 0000 = %05X   %05X\n", rdBank[0], wrBank[0]);
		ret += String.format("PAGE 4000 = %05X   %05X\n", rdBank[1], wrBank[1]);
		ret += String.format("PAGE 8000 = %05X   %05X\n", rdBank[2], wrBank[2]);
		ret += String.format("PAGE C000 = %05X   %05X\n", rdBank[3], wrBank[3]);
		ret += '\n';
		ret += mon.dumpDebug();
		if (h17 != null) ret += h17.dumpDebug();
		return ret;
	}

	private int read(boolean rom, int address) {
		if (rom && address < romMask) {
			if (address < monMask || (h17 == null && address >= 0x1800)) {
				return mon.read(address);
			}
			if (address >= 0x1800 && h17 != null) {
				return h17.read(address - 0x1800);
			}
		}
		return mem[address] & 0xff;
	}
	public int read(boolean rom, int bank, int address) {
		return read(rom, ((bank & 7) << 16) | address);
	}
	public int read(int address) {
		boolean _rom = rom;
		if (map) {
			// TODO: memory hole needed...
			if (!rom || address >= romMask ||
					(address >= monMask && address < 0x1800)) {
				_rom = false;
				int pg = (address >> 14) & 0x0003;
				address = rdBank[pg] + (address & 0x3fff);
			}
		}
		return read(_rom, address);
	}
	public void write(int address, int value) {
		if (rom && address < monMask) mon.write(address, value);
		if (map) {
			int pg = (address >> 14) & 0x0003;
			address &= 0x3fff;
			mem[wrBank[pg] + address] = (byte)value;
		} else {
			mem[address] = (byte)value;
		}
	}

	public void reset() {
		rom = true;	// TODO: let gpp do this?
		map = false;
		// NOTE: RESET does not affect 74LS670's contents
		//Arrays.fill(rdBank, 0);
		//Arrays.fill(wrBank, 0);
	}

	public void dumpCore(String file) {
		try {
			OutputStream core = new FileOutputStream(file);
			core.write(mem);
			core.close();
		} catch (Exception ee) {
			H89Operator.error(null, "Core Dump", ee.getMessage());
		}
	}

	public int interestedBits() { return gpioBit; }
	public void gppNewValue(int gpio) {
		rom = ((gpio & gpioORG0) == 0);
		if (romBit != 0) romGpp(gpio);
	}
}
