// Copyright (c) 2019 Douglas Miller <durgadas311@gmail.com>

import java.util.Arrays;
import java.util.Vector;
import java.util.Properties;
import java.io.*;
import java.lang.reflect.Constructor;

public class INS8250 implements IODevice, VirtualUART {
	static final int fifoLimit = 10; // should never even exceed 2
	private Interruptor intr;
	private int irq;
	private int src;
	private int basePort;
	private String name = null;
	private String prefix;
	private java.util.concurrent.LinkedBlockingDeque<Integer> fifo;
	private java.util.concurrent.LinkedBlockingDeque<Integer> fifi;
	private int DLL;
	private int DLH;
	private int IER;
	private int IIR;
	private int LCR;
	private int MCR;
	private int LSR;
	private int MSR;
	private int SCR;
	private int modem;

	private int clock16x = 1843200 / 16; // assume std 1.8432 MHz crystal

	private static final int IIR_NIP = 0x01; // No intr pending
	private static final int IIR_RXE = 0x06; // bits for Rx Error intr
	private static final int IIR_RXR = 0x04; // bits for Rx Data intr
	private static final int IIR_TXR = 0x02; // bits for Tx Empty intr
	private static final int IIR_MDM = 0x00; // bits for Modem Sts intr

	private static final int LCR_WLSM = 0x03;
	private static final int LCR_STB = 0x04;
	private static final int LCR_PEN = 0x08;
	private static final int LCR_DLAB = 0x80;

	private static final int IER_RXR = 0x01;
	private static final int IER_TXE = 0x02;
	private static final int IER_RXE = 0x04;
	private static final int IER_MDM = 0x08;

	private static final int LSR_RXR = 0x01;
	private static final int LSR_OVR = 0x02;
	private static final int LSR_PE = 0x04;
	private static final int LSR_FE = 0x08;
	private static final int LSR_BRK = 0x10;
	private static final int LSR_THE = 0x20;
	private static final int LSR_TSE = 0x40;
	private static final int LSR_RXERR = (LSR_BRK|LSR_FE|LSR_PE|LSR_OVR);

	private static final int MSR_CTS = 0x10;
	private static final int MSR_DSR = 0x20;
	private static final int MSR_RI  = 0x40;
	private static final int MSR_DCD = 0x80;

	private static final int MCR_DTR = 0x01;
	private static final int MCR_RTS = 0x02;
	private static final int MCR_OT1 = 0x04;
	private static final int MCR_OT2 = 0x08;

	private Object attObj;
	private OutputStream attFile;
	private SerialDevice io;
	private boolean io_in = false;
	private boolean io_out = false;
	private boolean excl = true;
	private long lastTx = 0;
	private long lastRx = 0;
	private long nanoBaud = 0; // length of char in nanoseconds
	private int bits; // bits per character
	private boolean c2550;	// TL16C2550 DUART variation?
	// TL16C2550 DUART deviates from INS8250/16550 by making
	// MCR.OUT2 also a master interrupt enable bit.
	// (no other known differences)

	public INS8250(Properties props, boolean h8_4, String pfx, int base, int irq,
			Interruptor intr) {
		name = pfx.toUpperCase() + "_INS8250";
		attObj = null;
		attFile = null;
		io = null;
		this.intr = intr;
		String s = props.getProperty(pfx + "_intr");
		if (s != null) {
			int i = Integer.valueOf(s);
			if (i >= 3 && i <= (h8_4 ? 7 : 5)) {
				irq = i;
			}
		}
		this.irq = irq;
		src = intr.registerINT(irq);
		if (h8_4) {
			// TODO: support "disabled"?
			s = props.getProperty(pfx + "_port");
			if (s != null) {
				int i = Integer.decode(s);
				if (i >= 0 && i <= 255 && (i & 7) == 0) {
					base = i;
				}
			}
		}
		basePort = base;
		s = props.getProperty(pfx + "_chip");
		c2550 = (s != null && s.equals("TL16C2550"));
		fifo = new java.util.concurrent.LinkedBlockingDeque<Integer>();
		fifi = new java.util.concurrent.LinkedBlockingDeque<Integer>();
		MSR = 0;
		reset();
		s = props.getProperty(pfx + "_att");
		if (s != null && s.length() > 1) {
			if (s.charAt(0) == '>') { // redirect output to file
				attachFile(s.substring(1));
			} else if (s.charAt(0) == '!') { // pipe to/from program
				attachPipe(s.substring(1));
			} else {
				attachClass(props, s);
			}
		}
	}

	private void attachFile(String s) {
		String[] args = s.split("\\s");
		setupFile(args, 0);
	}

	private void setupFile(String[] args, int start) {
		boolean append = false;
		for (int x = start + 1; x < args.length; ++x) {
			if (args[x].equals("a")) {
				append = true;
			} else if (args[x].equals("+")) {
				excl = false;
			}
		}
		if (args[start].equalsIgnoreCase("syslog")) {
			attFile = System.err;
			excl = false;
		} else try {
			attFile = new FileOutputStream(args[start], append);
			if (excl) {
				setModem(VirtualUART.SET_CTS | VirtualUART.SET_DSR);
			}
		} catch (Exception ee) {
			System.err.format("Invalid file in attachment: %s\n", args[start]);
			return;
		}
	}

	private void attachPipe(String s) {
		System.err.format("Pipe attachments not yet implemented: %s\n", s);
	}

	private void attachClass(Properties props, String s) {
		String[] args = s.split("\\s");
		for (int x = 1; x < args.length; ++x) {
			if (args[x].startsWith(">")) {
				excl = false;
				args[x] = args[x].substring(1);
				setupFile(args, x);
				// TODO: truncate args so Class doesn't see?
			}
		}
		Vector<String> argv = new Vector<String>(Arrays.asList(args));
		// try to construct from class...
		try {
			Class<?> clazz = Class.forName(args[0]);
			Constructor<?> ctor = clazz.getConstructor(
					Properties.class,
					argv.getClass(),
					VirtualUART.class);
			// funky "new" avoids "argument type mismatch"...
			attObj = ctor.newInstance(
					props,
					argv,
					(VirtualUART)this);
		} catch (Exception ee) {
			System.err.format("Invalid class in attachment: %s\n", s);
			return;
		}
	}

	public void attachDevice(SerialDevice io) {
		this.io = io;
		io_in = (io != null && (io.dir() & SerialDevice.DIR_IN) != 0);
		io_out = (io != null && (io.dir() & SerialDevice.DIR_OUT) != 0);
		modem = -1;
		changeModem();
	}

	public void setClock(int clk) {
		clock16x = clk / 16;
	}

	public void rcvFlush() {
		fifi.clear();
		chkIntr();
	}

	public void sndFlush() {
		fifo.clear();
		chkIntr();
	}

	public int sndSize() {
		return fifo.size();
	}

	public boolean rcvAll(java.util.concurrent.LinkedBlockingDeque<Integer> dst,
			int tmp) {
		int did = 0;
		try {
			// TODO: might require LSR events, too
			while (fifi.size() > 0) {
				dst.add(fifi.take() | tmp);
				++did;
			}
		} catch (Exception ee) {}
		if (fifi.size() == 0) {
			LSR &= ~LSR_RXR;
			chkIntr();
		}
		return (did > 0);
	}

	public void sndOne(int val) {
		if (attFile != null) {
			try {
				attFile.write(val);
			} catch (Exception ee) {}
		}
		if (io_out) {
			io.write(val);
		}
		if ((attFile == null && !io_out) || !excl) {
			fifo.add(val); // don't limit size
			LSR &= ~LSR_TSE;
			if (fifo.size() > 1) {
				LSR &= ~LSR_THE;
			}
			chkIntr();
		}
	}

	// Conditions affecting interrupts have changed, ensure proper signal.
	private void chkIntr() {
		int bits = IIR_NIP; // No Intr
		if ((IER & IER_RXE) != 0 && (LSR & LSR_RXERR) != 0) {
			// Rx Errors never happen... almost never
			bits = IIR_RXE;
		} else if ((IER & IER_RXR) != 0 && (LSR & LSR_RXR) != 0) {
			bits = IIR_RXR;
		} else if ((IER & IER_TXE) != 0 && false) {
			// Tx always appears empty... So never transitions...
			// TODO: must raise INT sometime...
			bits = IIR_TXR;
		} else if ((IER & IER_TXE) != 0 && (MSR & 0x0f) != 0) {
			bits = IIR_MDM;
		}
		if (c2550 && (MCR & MCR_OT2) == 0) {
			bits = IIR_NIP;
		}
		// TODO: only raise/lower if conditions change
		IIR = bits;
		if ((IIR & IIR_NIP) == 0) {
			intr.raiseINT(irq, src);
		} else {
			intr.lowerINT(irq, src);
		}
	}

	///////////////////////////////
	/// Interfaces for IODevice ///
	public int in(int port) {
		int off = port - basePort;
		int val = 0;
		switch(off) {
		case 0: // Rx Data
			if ((LCR & LCR_DLAB) != 0) {
				val = DLL;
				break;
			}
			synchronized(this) {
				if (fifi.size() > 0) {
					try {
						val = fifi.take();
					} catch (Exception ee) {}
					if (fifi.size() == 0) {
						LSR &= ~LSR_RXR;
						chkIntr();
					}
				}
			}
			break;
		case 1:
			if ((LCR & LCR_DLAB) == 0) {
				val = IER;
			} else {
				val = DLH;
			}
			break;
		case 2: // IIR
			val = IIR;
			break;
		case 3:
			val = LCR;
			break;
		case 4:
			val = MCR;
			break;
		case 5:
			long t = System.nanoTime();
			if (t - lastTx > nanoBaud) {
				if (attFile != null || io_out) {
					LSR |= LSR_THE;
					LSR |= LSR_TSE;
					lastTx = t;
				} else if (fifo.size() < 2) {
					LSR |= LSR_THE;
					if (fifo.size() < 1) {
						LSR |= LSR_TSE;
					} else {
						lastTx = t;
					}
				}
			}
			// simulate special case in MMS 84B baud test.
			// 444-84B intentionally lets RX overrun by 50mS delay,
			// but that delay seems to not always be exactly 50mS.
			// It seems that the 2mS clock "backlog" can result
			// in inaccuracies at this scale.
			if (io_in) {
				if (io.available() > 0) {
					LSR |= LSR_RXR;
				}
			} else {
				while (t - lastRx > 30000000 && fifi.size() > 1) {
					try {
						fifi.take();
					} catch (Exception ee) {}
					LSR |= LSR_OVR;
				}
				lastRx = t;
			}
			val = LSR;
			// This resets overrun, framing, and parity errors,
			// but those are never set. Almost never.
			LSR &= ~LSR_RXERR;
			chkIntr();
			break;
		case 6:
			val = MSR;
			MSR &= 0xf0;
			chkIntr();
			break;
		case 7:
			val = SCR;
			break;
		}
		return val;
	}

	public void out(int port, int val) {
		int off = port - basePort;
		val &= 0xff; // necessary?
		switch(off) {
		case 0: // Tx Data
			if ((LCR & LCR_DLAB) != 0) {
				DLL = val;
				break;
			}
			// discard char if not initialized,
			// TODO: still log char to attFile?
			if (nanoBaud == 0) {
				break;
			}
			if (attFile != null) {
				try {
					attFile.write(val);
				} catch (Exception ee) {}
			}
			if (io_out) {
				io.write(val);
			}
			if ((attFile == null && !io_out) || !excl) {
				fifo.add(val);
				lastTx = System.nanoTime();
				LSR &= ~LSR_TSE;
				if (fifo.size() > 1) {
					LSR &= ~LSR_THE;
					try {
						while (fifo.size() > fifoLimit) {
							fifo.removeFirst();
						}
					} catch (Exception ee) {}
				} else {
					// probably already set
					LSR |= LSR_THE;
				}
			}
			// TODO: turn off TXE intr, but must turn on later.
			// LSR &= ~LSR_TSE; // or THE?
			// chkIntr();
			break;
		case 1:
			if ((LCR & LCR_DLAB) == 0) {
				IER = val;
				chkIntr();
			} else {
				DLH = val;
			}
			break;
		case 2: // IIR
			// IIR = val; // Writing not allowed...
			break;
		case 3:
			int diff = LCR ^ val;
			LCR = val;
			if ((diff & (LCR_WLSM | LCR_STB | LCR_PEN)) != 0) {
				// force recomputing of baud (if DLAB == 0)
				diff |= LCR_DLAB;
				bits = 5 + (LCR & LCR_WLSM) +
					((LCR & LCR_STB) == 0 ? 1 : 2) +
					((LCR & LCR_PEN) == 0 ? 0 : 1);
			}
			if ((diff & LCR_DLAB) != 0 && (LCR & LCR_DLAB) == 0) {
				// update Baud, even if it was only being read.
				if (DLH == 0 && DLL == 0) {
					nanoBaud = 0;
				} else {
					int baud = clock16x / ((DLH << 8) | DLL);
					int cps = baud / bits;
					nanoBaud = (1000000000 / cps);
				}
			}
			break;
		case 4:
			MCR = val;
			changeModem();
			break;
		case 5:
			// LSR = val; // Writing not allowed...
			break;
		case 6:
			// MSR = val; // Writing not allowed...
			break;
		case 7:
			SCR = val;
			break;
		}
	}

	public void reset() {
		DLL = 0;
		DLH = 0;
		IER = 0;
		IIR = IIR_NIP;
		intr.lowerINT(irq, src);
		LCR = 0;
		MCR = 0;
		modem = -1;
		LSR = (LSR_THE | LSR_TSE);
		// MSR = 0; // do not touch MSR, it is set by external entities
		SCR = 0;
		fifo.clear();
		fifi.clear();
		changeModem();
	}
	public int getBaseAddress() {
		return basePort;
	}
	public int getNumPorts() {
		return 8;
	}

	private void changeModem() {
		if (io == null) {
			return;
		}
		int m = getModem();
		int diff = modem ^ m;
		if (diff == 0) {
			return;
		}
		modem = m;
		io.modemChange(this, modem);
	}

	////////////////////////////////////////////////////
	/// Interfaces for the virtual peripheral device ///
	public boolean attach(Object periph) { return false; }
	public void detach() {
		if (attObj == null) {
			return;
		}
		attObj = null;
		excl = true;
		try {
			fifo.addFirst(-1);
		} catch (Exception ee) {
			fifo.add(-1);
		}
	}
	public int available() {
		return fifo.size();
	}

	// Must sleep if nothing available...
	public int take() {
		try {
			int c = fifo.take();
			// Tx always appears empty...
			// But might need to simulate intr.
			// This is separate thread from CPU so must be careful...
			// TBD: LSR |= LSR_THE; chkIntr();
			return c;
		} catch(Exception ee) {
			return -1;
		}
	}

	public boolean ready() {
		return (LSR & LSR_RXR) == 0;
	}
	// Must NOT sleep
	public synchronized void put(int ch, boolean sleep) {
		// TODO: prevent infinite growth?
		fifi.add(ch & 0xff);
		lastRx = System.nanoTime();
		LSR |= LSR_RXR;
		chkIntr();
	}

	public void setModem(int mdm) {
		int old = MSR;
		int nuw = 0;
		if ((mdm & VirtualUART.SET_CTS) != 0) {
			nuw |= MSR_CTS;
		}
		if ((mdm & VirtualUART.SET_DSR) != 0) {
			nuw |= MSR_DSR;
		}
		if ((mdm & VirtualUART.SET_DCD) != 0) {
			nuw |= MSR_DCD;
		}
		if ((mdm & VirtualUART.SET_RI) != 0) {
			nuw |= MSR_RI;
		}
		old = ((old ^ nuw) >> 4) & 0x0f;
		MSR = nuw | old;
		// TODO: must make this thread-safe...
		chkIntr();
		changeModem();
	}
	public int getModem() {
		int mdm = 0;
		if ((MCR & MCR_DTR) != 0) {
			mdm |= VirtualUART.GET_DTR;
		}
		if ((MCR & MCR_RTS) != 0) {
			mdm |= VirtualUART.GET_RTS;
		}
		if ((MCR & MCR_OT1) != 0) {
			mdm |= VirtualUART.GET_OT1;
		}
		if ((MCR & MCR_OT2) != 0) {
			mdm |= VirtualUART.GET_OT2;
		}
		if ((MSR & MSR_CTS) != 0) {
			mdm |= VirtualUART.SET_CTS;
		}
		if ((MSR & MSR_DSR) != 0) {
			mdm |= VirtualUART.SET_DSR;
		}
		if ((MSR & MSR_DCD) != 0) {
			mdm |= VirtualUART.SET_DCD;
		}
		if ((MSR & MSR_RI) != 0) {
			mdm |= VirtualUART.SET_RI;
		}
		return mdm;
	}
	public String getPortId() { return prefix; }

	public String getDeviceName() { return name; }

	public String dumpDebug() {
		String ret = new String();
		ret += String.format("port %02x, #fifo = %d, #fifi = %d\n",
			basePort, fifo.size(), fifi.size());
		ret += String.format("DLL = %02x, DLH = %02x\n", DLL, DLH);
		ret += String.format("IER = %02x, IIR = %02x\n", IER, IIR);
		ret += String.format("LCR = %02x, LSR = %02x\n", LCR, LSR);
		ret += String.format("MCR = %02x, MSR = %02x\n", MCR, MSR);
		ret += String.format("nanoBaud = %d, clock16x = %d\n",
				nanoBaud, clock16x);
		if (io != null) {
			ret += io.dumpDebug();
		}
		return ret;
	}
}
