// Copyright (c) 2016 Douglas Miller <durgadas311@gmail.com>

import java.util.Arrays;
import java.util.Vector;
import java.util.Properties;
import java.awt.event.*;
import javax.swing.*;

public class MMS77316 extends WD1797
		implements DiskController, InterruptController, WD1797Controller,
			ActionListener {
	// TODO: These need to come from CPU (Z80)...
	static final int EI = 0xfb;
	static final int RST6 = 0xf7;

	static final int BasePort_c = 0x38;
	static final int Wd1797_Offset_c = 4;
	static final int ControlPort_Offset_c = 0;
	static final int MMS77316_NumPorts_c = 8;
	static final String MMS77316_Name_c = "MMS77316";
	static final int numDisks_c = 8;
	// NOTE! we use INT5 signal but force RST6 on data bus!
	static final int MMS77316_Intr_c = 5; // IRQ5

	static final int ctrl_EnableIntReq_c = 0x08;
	static final int ctrl_EnableBurstN_c = 0x20;
	static final int ctrl_SetMFMRecordingN_c = 0x40;
	static final int ctrl_DriveSel_c = 0x07;
	static final int ctrl_525DriveSel_c = 0x04;

	private GenericFloppyDrive[] drives_m;
	private LED[] leds_m;
	private int srcDrq;
	private int srcIntrq;
	private Interruptor intr;
	private javax.swing.Timer timer;

	private int controlReg_m;
	private int drqCount_m;

	public MMS77316(Properties props, String slot, LEDHandler lh, Interruptor intr) {
		super(BasePort_c + Wd1797_Offset_c, intr);
		super.setController(this);
		this.intr = intr;
		timer = new javax.swing.Timer(2000, this);
		//timer.setRepeats(false);
		srcDrq = intr.registerINT(MMS77316_Intr_c);
		srcIntrq = intr.registerINT(MMS77316_Intr_c);
		intr.addIntrController(this);
		leds_m = new LED[numDisks_c];
		drives_m = new GenericFloppyDrive[numDisks_c];
		String s;
		Arrays.fill(drives_m, null);
		controlReg_m = 0;
		drqCount_m = 0;

		// First identify what drives are installed.
		GenericFloppyDrive drv;
		for (int x = 0; x < numDisks_c; ++x) {
			String prop = String.format("mms77316_drive%d", x + 1);
			s = props.getProperty(prop);
			if (s != null) {
				drv = installDrive(x, s, lh);
			}
		}

		// Next identify what diskettes are pre-inserted.
		for (int x = 0; x < numDisks_c; ++x) {
			String prop = String.format("mms77316_disk%d", x + 1);
			s = props.getProperty(prop);
			if (s != null) {
				drv = getDrive(x);
				if (drv != null) {
					drv.insertMedia(s.split("\\s"));
				}
			}
		}
		reset();
	}

	private GenericFloppyDrive installDrive(int x, String s, LEDHandler lh) {
		String[] ss = s.split("\\s", 2);
		String n = getDriveName(x);
		if (ss.length >= 2) {
			n = ss[1];
		}
		GenericFloppyDrive drv = GenericFloppyDrive.getInstance(ss[0], n, lh);
		if (!connectDrive(x, drv)) {
			return null;
		}
		leds_m[x] = drv.getLED();
		return drv;
	}

	public int getBaseAddress() { return BasePort_c; }
	public int getNumPorts() { return MMS77316_NumPorts_c; }

	private void setCtrlReg(int val) {
		int prev = controlReg_m & ctrl_DriveSel_c;
		int diff = controlReg_m ^ val;
		controlReg_m = val;
		if ((diff & ctrl_EnableIntReq_c) != 0 &&
				(val & ctrl_EnableIntReq_c) == 0) {
			// This disables any drive select...
			if (leds_m[prev] != null) {
				leds_m[prev].set(false);
			}
			//System.err.format("Drive select OFF\n");
		} else if ((diff & ctrl_DriveSel_c) != 0) {
			// TODO: only if ctrl_EnableIntReq_c == 1?
			// change drive select LED...
			if (leds_m[prev] != null) {
				leds_m[prev].set(false);
			}
			int next = controlReg_m & ctrl_DriveSel_c;
			if (leds_m[next] != null) {
				leds_m[next].set(true);
			}
			//System.err.format("Drive select %d\n", val & ctrl_DriveSel_c);
		}
	}

	public GenericFloppyDrive getCurDrive() {
		// This could be null
		return drives_m[controlReg_m & ctrl_DriveSel_c];
	}

	public int getClockPeriod() {
		return (controlReg_m & ctrl_525DriveSel_c) != 0 ? 1000 : 500;
	}

	public Vector<GenericRemovableDrive> getDiskDrives() {
		Vector<GenericRemovableDrive> drives = new Vector<GenericRemovableDrive>();
		for (int x = 0; x < numDisks_c; ++x) {
			drives.add(drives_m[x]);	// might be null, but must preserve number.
		}
		return drives;
	}

	public int getDriveSize(int index) {
		if (index < 0 || index >= numDisks_c) {
			return 0;
		}
		if ((index & ctrl_525DriveSel_c) != 0) {
			return 5;
		} else {
			return 8;
		}
	}

	public String getDriveName(int index) {
		if (index < 0 || index >= numDisks_c) {
			return null;
		}
		String str = String.format("%s-%d", MMS77316_Name_c, index + 1);
		return str;
	}

	public String getDeviceName() { return MMS77316_Name_c; }

	public GenericDiskDrive findDrive(String ident) {
		for (int x = 0; x < numDisks_c; ++x) {
			if (drives_m[x] == null) {
				continue;
			}
			if (ident.equals(drives_m[x].getDriveName())) {
				return drives_m[x];
			}
		}
		return null;
	}

	public int readDataBus() {
		int opCode = -1;
		// The MMS77316 controller assumes it is the highest priority interrupt,
		// so if we have an interrupt asserted then we take control.
		if (intrqAllowed() && (intrqRaised_m || drqRaised_m)) {
			// TODO: generate op-codes without knowing Z80CPU?
			opCode = (intrqRaised_m ? RST6 : EI);
			// TODO: should anything be reset here?
			//intrqRaised_m = false;
		}
		return opCode;
	}

	public void destroy() {
		for (int x = 0; x < numDisks_c; ++x) {
			if (drives_m[x] != null) {
				drives_m[x].insertMedia(null);
			}
		}
	}

	public void reset() {
		super.reset();
		setCtrlReg(0);
		intr.lowerINT(MMS77316_Intr_c, srcDrq);
		intr.lowerINT(MMS77316_Intr_c, srcIntrq);
	}

	public int in(int addr) {
		int offset = addr - BasePort_c;
		int val = 0;
		// case ControlPort_Offset_c: NOT READABLE
		if (offset >= Wd1797_Offset_c) {
			if (offset - Wd1797_Offset_c == DataPort_Offset_c) {
				// might need to simulate WAIT states...
				// Must NOT wait too long - this blocks all other progress.
				// TODO: redesign this to return to execute() loop and
				// stall there... requires in() to return a status or
				// some other way inform the CPU to stall. The MMS77316
				// has a built-in timeout on the WAIT hardware anyway,
				// so, insure we don't stay here forever. The hardware
				// timed out after 16 busclk (2MHz, i.e. CPU clock) cycles,
				// really should count those but this is probably close enough.
				int timeout = 0;
				while (burstMode() && !drqRaised_m && !intrqRaised_m && ++timeout < 16) {
					intr.waitCPU();
				}
			}
			val = super.in(addr);
		} else {
			System.err.format("MMS77316::in(Unknown - 0x%02x)\n", addr);
		}
		return val;
	}

	public void out(int addr, int val) {
		int offset = addr - BasePort_c;
		if (offset >= Wd1797_Offset_c) {
			if (offset - Wd1797_Offset_c == DataPort_Offset_c) {
				// See notes for MMS77316::in()...
				int timeout = 0;
				while (burstMode() && !drqRaised_m && !intrqRaised_m && ++timeout < 16) {
					intr.waitCPU();
				}
			}
			super.out(addr, val);
		} else if (offset == ControlPort_Offset_c) {
			setCtrlReg(val);
			drqCount_m = 0;
			if ((controlReg_m & ctrl_525DriveSel_c) != 0) {
				GenericFloppyDrive drive = getCurDrive();
				if (drive != null) {
					drive.motor(true);
				}
			}
			if ((controlReg_m & ctrl_EnableIntReq_c) != 0) {
				// Interrupts now unmasked, so raise as appropriate.
				if (intrqRaised_m) {
					intr.raiseINT(MMS77316_Intr_c, srcIntrq);
				}
				if (drqRaised_m) {
					intr.raiseINT(MMS77316_Intr_c, srcDrq);
				}
			}
		} else {
			System.err.format("MMS77316::out(Unknown - 0x%02x): %d\n", addr, val);
		}
	}

	public GenericFloppyDrive getDrive(int unitNum) {
		if (unitNum < numDisks_c) {
			return drives_m[unitNum];
		}
		return null;
	}

	public boolean connectDrive(int unitNum, GenericFloppyDrive drive) {
		if (unitNum >= numDisks_c) {
			System.err.format("MMS77316 Invalid unit number (%d)\n", unitNum);
			return false;
		}
		if (drives_m[unitNum] != null) {
			System.err.format("MMS77316 %d: drive already connect\n", unitNum);
			return false;
		}
		if (getDriveSize(unitNum) != drive.getMediaSize()) {
			System.err.format("MMS77316 %d: drive incompatible\n", unitNum);
			return false;
		}
		drives_m[unitNum] = drive;
		return true;
	}

	public boolean removeDrive(int unitNum) {
		return false;
	}

	public void raisedIntrq() {
		drqCount_m = 0;
		if (intrqAllowed()) {
			intr.raiseINT(MMS77316_Intr_c, srcIntrq);
			return;
		}
	}

	public void raisedDrq() {
		if (drqAllowed()) {
			++drqCount_m;
			intr.raiseINT(MMS77316_Intr_c, srcDrq);
		}
	}

	public void loweredIntrq() {
		intr.lowerINT(MMS77316_Intr_c, srcIntrq);
	}

	public void loweredDrq() {
		intr.lowerINT(MMS77316_Intr_c, srcDrq);
	}

	public void loadedHead(boolean load) {
		if (load) {
			// deviation from hardware? trigger clear from
			// start of HLD, not end (driver never unloads head?)
			timer.removeActionListener(this);
			timer.addActionListener(this);
			timer.restart();
		} else {
			//timer.removeActionListener(this);
			//timer.addActionListener(this);
			//timer.restart();
			// TODO: start another timer for motor-off (if was on...)?
		}
	}

	private boolean burstMode() {
		return (controlReg_m & ctrl_EnableBurstN_c) == 0;
	}
	private boolean intrqAllowed() {
		return (controlReg_m & ctrl_EnableIntReq_c) != 0;
	}
	private boolean drqAllowed() {
		return (controlReg_m & ctrl_EnableIntReq_c) != 0 &&
			((controlReg_m & ctrl_EnableBurstN_c) != 0 || drqCount_m < 1);
	}
	public boolean doubleDensity() {
		return (controlReg_m & ctrl_SetMFMRecordingN_c) == 0;
	}

	public void actionPerformed(ActionEvent e) {
		if (e.getSource() == timer) {
			timer.removeActionListener(this);
			setCtrlReg(0);
			// TODO: turn off motor, too?
			//GenericFloppyDrive drive = getCurDrive();
			//if (drive != null) {
			//	drive.motor(true);
			//}
		}
	}

	public String dumpDebug() {
		String ret = String.format(
			"CTRL=%02x\n" +
			"FDC-STS=%02x FDC-CMD=%02x\n" +
			"FDC-TRK=%d FDC-SEC=%d FDC-DAT=%02x\n" +
			"DRQ=%s INTRQ=%s HLD=%s\n",
			controlReg_m,
			statusReg_m, cmdReg_m,
			trackReg_m, sectorReg_m, dataReg_m,
			Boolean.toString(drqRaised_m), Boolean.toString(intrqRaised_m),
			Boolean.toString(headLoaded_m));
		return ret;
	}
}
