// Copyright (c) 2021 Douglas Miller <durgadas311@gmail.com>

import java.util.Arrays;
import java.util.Properties;
import java.util.Vector;
import java.awt.event.*;
import java.io.*;

// This uses the CPM3.SYS file format, although code could be
// almost anything. Should be possible to put a CP/M 2.2 image
// in this format as comlen, and leave bnklen zero.
//
// NOTE: complications arise for OS versions that require
// files to complete booting, so truly diskless (CP/NOS) is
// not yet supported. Right now, a diskfull OS image could
// be booted over the network, provided the necessary files
// are still accessible once the OS is initialized. The merits
// of such a capability are debatable.
//
// For both CP/NET 1.2 and CP/NET 3, the OS must be fully
// initialized in order to load/start NDOS and SNIOS (CP/NET).
// Once we pass control to the CBOOT entry, we lose the
// ability to influence what gets executed. We can't ever
// call CBOOT once NDOS/SNIOS are running or those get
// obliterated.

// Message format:
//    LN str... 00
// where LN is length of string, including NUL (00).

public class NetBootSys implements NetBoot {
	// CPM3.SYS file format:
	static final int MEMTOP = 0;
	static final int COMLEN = 1;
	static final int BNKTOP = 2;
	static final int BNKLEN = 3;
	static final int ENTRY = 4;	// 2 bytes
	static final int CNFTBL = 6;	// 2 bytes
	static final int COPY = 16;	// DRI Copyright message - use to detect ORG0
	static final int SIGNON = 128;	// '$' terminated string, 128 bytes max
	static final int IMAGE = 256;	// start of OS image (end header)

	static final int RECLEN = 128;

	private int sid;
	private int nid;

	private int memtop;
	private int comlen;
	private int bnktop;
	private int bnklen;
	private int entry;
	private int cfgtbl;
	private int curdma;
	private int reccnt;
	private boolean org0;
	private byte[] image;

	public NetBootSys(String dir, String def, byte[] msg) throws Exception {
		nid = msg[NetworkServer.SID] & 0xff;
		sid = msg[NetworkServer.DID] & 0xff;
		int n = (msg[NetworkServer.SIZ] & 0xff) + 1;	// first byte LEN, last NUL
		String tag = "";
		if (n > 2) {
			tag = new String(msg, NetworkServer.DAT + 1, n - 2).toLowerCase();
		}
		File file = null;
		if (tag.length() > 0) {
			// search for a variation of filename that exists...
			file = new File(String.format("%s/cid%02X%s", dir, nid, tag));
//System.err.format("Trying %s\n", file.getAbsolutePath());
			if (!file.exists()) {
				file = new File(String.format("%s/cid%02X%s.sys", dir, nid, tag));
//System.err.format("Trying %s\n", file.getAbsolutePath());
			}
			if (!file.exists()) {
				file = new File(String.format("%s/%s", dir, tag));
//System.err.format("Trying %s\n", file.getAbsolutePath());
			}
			if (!file.exists()) {
				file = new File(String.format("%s/%s.sys", dir, tag));
//System.err.format("Trying %s\n", file.getAbsolutePath());
			}
		} else {
			file = new File(String.format("%s/cid%02X.sys", dir, nid));
//System.err.format("Trying %s\n", file.getAbsolutePath());
			if (!file.exists()) {
				file = new File(String.format("%s/%s", dir, def));
//System.err.format("Trying %s\n", file.getAbsolutePath());
			}
		}
		if (!file.exists()) {
System.err.format("No boot file \"%s\"\n", file.getAbsolutePath());
			throw new FileNotFoundException(file.getAbsolutePath());
		}
		try {
			FileInputStream f = new FileInputStream(file);
			image = new byte[f.available()];
			f.read(image);
			f.close();
		} catch (Exception ee) {
ee.printStackTrace();
			throw ee;
		}
		if (image[1] == 0 && image[3] == 0) {
System.err.format("No segments to load \"%s\"\n", file.getAbsolutePath());
			throw new RuntimeException("No segments to load");
		}
		memtop = (image[MEMTOP] & 0xff) << 8;	// 0000 mean 10000
		comlen = (image[COMLEN] & 0xff) << 8;
		bnktop = (image[BNKTOP] & 0xff) << 8;
		bnklen = (image[BNKLEN] & 0xff) << 8;
		entry = (image[ENTRY] & 0xff) | ((image[ENTRY+1] & 0xff) << 8);
		cfgtbl = (image[CNFTBL] & 0xff) | ((image[CNFTBL+1] & 0xff) << 8);
		if (cfgtbl != 0) {
			fixCfgTbl();
		}
		org0 = (image[COPY] == 'C');
		System.err.format("Client %02x booting file \"%s\" (ORG0 %s)\n",
			nid, file.getAbsolutePath(), org0);
	}

	private void fixCfgTbl() {
		// cfgtbl must be within RES segment, and must not cross
		// record boundary.
		// copute distance from top - i.e. start of image
		int off = (memtop - (cfgtbl & 0xff80) - 128) & 0xffff;
		off |= (cfgtbl & 0x007f);
		off += IMAGE + 2;	// start of drive table
		// fixup A:-P:,CON:,LST:
		for (int x = 0; x < 16+2; ++x) {
			int i = x * 2;
			if ((image[off + i] & 0x80) != 0 &&
				(image[off + i + 1] & 0xff) == 0xff) {
				image[off + i + 1] = (byte)(sid);
			}
		}
	}

	public boolean hasRes() { return (comlen > 0); }
	public boolean hasBnk() { return (bnklen > 0); }

	public void setLoadMsg(byte[] resp, int start, int max) {
		System.arraycopy(image, 128, resp, start, 128);
	}

	public int startResAddr() {
		curdma = comlen + IMAGE - RECLEN;
		reccnt = comlen / RECLEN;
		return (memtop - comlen) & 0xffff;
	}

	public int nextResRec(byte[] resp, int start, int max) {
		System.arraycopy(image, curdma, resp, NetworkServer.DAT, RECLEN);
		curdma -= RECLEN;
		--reccnt;
		return reccnt;
	}

	public int startBnkAddr() {
		curdma = (comlen + bnklen) + IMAGE - RECLEN;
		reccnt = bnklen / RECLEN;
		return (bnktop - bnklen) & 0xffff;
	}

	public int nextBnkRec(byte[] resp, int start, int max) {
		System.arraycopy(image, curdma, resp, NetworkServer.DAT, RECLEN);
		curdma -= RECLEN;
		--reccnt;
		return reccnt;
	}

	public int entryAddr() {
		return entry;
	}

	public boolean needsOrg0() {
		return org0;
	}
}
