// Copyright (c) 2019 Douglas Miller <durgadas311@gmail.com>

import java.util.Arrays;
import java.util.Properties;
import java.util.Vector;
import java.awt.event.*;
import java.io.*;

// This uses the CPM3.SYS file format, although code could be
// almost anything. Should be possible to put a CP/M 2.2 image
// in this format as comlen, and leave bnklen zero.
//
// NOTE: complications arise for OS versions that require
// files to complete booting, so truly diskless (CP/NOS) is
// not yet supported. Right now, a diskfull OS image could
// be booted over the network, provided the necessary files
// are still accessible once the OS is initialized. The merits
// of such a capability are debatable.
//
// For both CP/NET 1.2 and CP/NET 3, the OS must be fully
// initialized in order to load/start NDOS and SNIOS (CP/NET).
// Once we pass control to the CBOOT entry, we lose the
// ability to influence what gets executed. We can't ever
// call CBOOT once NDOS/SNIOS are running or those get
// obliterated.

// Basic semantics on Server:
//
// if msg[FMT] == NetworkBoot.FMT then
//     if nwb == null then
//         nwb = NetworkBoot.doStart(props, msg)
//         if nwb == null then
//             send(NetworkBoot.error(msg))
//             return
//         endif
//     endif
//     resetTimeout()
//     send(nwb.doNext(msg))
//     if nwb.done() then
//         nwb = null
//     endif
// endif
//
// timeout():
//     nwb = null
//

public class NetworkBoot {
	// route all messages with this FMT to us:
	public static final int FMT = 0xb0;

	static final int STATE_START = 0;
	static final int STATE_RESDMA = 1;
	static final int STATE_RESLD = 2;
	static final int STATE_BNKDMA = 3;
	static final int STATE_BNKLD = 4;
	static final int STATE_ORG0DMA = 5;
	static final int STATE_ORG0LD = 6;
	static final int STATE_DONE = 7;

	// CPM3.SYS file format:
	static final int MEMTOP = 0;
	static final int COMLEN = 1;
	static final int BNKTOP = 2;
	static final int BNKLEN = 3;
	static final int ENTRY = 4;	// 2 bytes
	static final int CNFTBL = 6;	// 2 bytes
	static final int COPY = 16;	// DRI Copyright message - use to detect ORG0
	static final int SIGNON = 128;	// '$' terminated string, 128 bytes max
	static final int IMAGE = 256;	// start of OS image (end header)

	static final int RECLEN = 128;

	private int state = -1;
	private byte[] image;
	private byte[] org0; // may be null
	private int org0Adr;
	private int curdma;
	private int reccnt;
	private NetBoot netBoot;

	private byte[] resp;

	public NetworkBoot(NetBoot nb, byte[] img0, int adr0, int sid) {
		netBoot = nb;
		org0 = img0;
		org0Adr = adr0;
		resp = new byte[128 + NetworkServer.DAT];
		state = STATE_START;
	}

	// If returns null, caller sends error(msg)
	public static NetworkBoot doStart(Properties props, byte[] msg) {
		// Can't get here unless FMT == B0
		if (msg[NetworkServer.FNC] != 1 &&
				msg[NetworkServer.FNC] != 2) {
			System.err.format("Invalid boot FNC %02x\n",
					msg[NetworkServer.FNC] & 0xff);
			return null;
		}
		// since dir is used read-only, not so tight on security.
		String dir = props.getProperty("netboot_dir");
		if (dir == null) {
			dir = System.getProperty("user.home") + "/NetBoot";
		}
		String org0 = props.getProperty("netboot_org0");
		if (org0 == null) {
			org0 = "org0boot.bin";
		} else if (org0.equalsIgnoreCase("none")) {
			org0 = null;
		}
		int adr0 = 0x3000;
		String s = props.getProperty("netboot_org0_adr");
		if (s != null) {
			adr0 = Integer.decode(s);
		}
		String def = props.getProperty("netboot_default");
		if (def == null) {
			def = "defboot.sys";
		}
		byte[] img0 = null;
		NetBoot nb = null;
		try {
			if (msg[NetworkServer.FNC] == 1) {
				nb = new NetBootSys(dir, def, msg);
			} else {
				nb = new NetBootLdr(dir, def, msg);
			}
		} catch (Exception ee) {
			ee.printStackTrace();
			return null;
		}
		if (org0 != null && nb.needsOrg0()) {
			File o0 = new File(dir + "/" + org0);
			// No need to check existence, this will catch it.
			try {
				FileInputStream f = new FileInputStream(o0);
				img0 = new byte[f.available()];
				f.read(img0);
				f.close();
			} catch (Exception ee) {
				ee.printStackTrace();
				return null;
			}
		}
		// TODO: if error...
		int sid = msg[NetworkServer.DID] & 0xff; // our ID
		return new NetworkBoot(nb, img0, adr0, sid);
	}

	public static byte[] error(byte[] msg) {
		return new byte[]{ (byte)0xb1, msg[NetworkServer.SID],
					msg[NetworkServer.DID], 0, 0, 0 };
	}

	public boolean done() {
		return (state < 0 || state > STATE_DONE);
	}

	// Never returns null.
	public byte[] doNext(byte[] msg) {
		// Can't get here unless FMT == B0
		int adr;
		int n;
		// TODO: validate SIZ, DAT?
		// TODO: FNC might be 1...
		int fnc = msg[NetworkServer.FNC] & 0xff;
		if ((state != STATE_START && fnc != 0) ||
			(state == STATE_START && (fnc != 1 && fnc != 2))) {
			state = -1;
		}
		switch (state) {
		case STATE_START:
			// TODO: does empty message require special?
			// TODO: scan for term, send only part used?
			resp[NetworkServer.FNC] = (byte)0x01;
			resp[NetworkServer.SIZ] = (byte)(128-1);
			netBoot.setLoadMsg(resp, NetworkServer.DAT, 128);
			if (netBoot.hasRes()) {
				state = STATE_RESDMA;
			} else {
				state = STATE_BNKDMA;
			}
			break;
		case STATE_RESDMA:	// start RES boot
			adr = netBoot.startResAddr();
			resp[NetworkServer.FNC] = (byte)0x02;
			resp[NetworkServer.SIZ] = (byte)(2-1);
			resp[NetworkServer.DAT] = (byte)adr;
			resp[NetworkServer.DAT + 1] = (byte)(adr >> 8);
			state = STATE_RESLD;
			break;
		case STATE_RESLD: // send RES records
			// don't land here unless comlen > 0 (1+ records)
			resp[NetworkServer.FNC] = (byte)0x03;
			resp[NetworkServer.SIZ] = (byte)(RECLEN-1);
			n = netBoot.nextResRec(resp, NetworkServer.DAT, RECLEN);
			if (n <= 0) {
				if (netBoot.hasBnk()) {
					state = STATE_BNKDMA;
				} else if (org0 != null) {
					state = STATE_ORG0DMA;
				} else {
					state = STATE_DONE;
				}
			}
			break;
		case STATE_BNKDMA:	// start BNK boot
			adr = netBoot.startBnkAddr();
			resp[NetworkServer.FNC] = (byte)0x02;
			resp[NetworkServer.SIZ] = (byte)(2-1);
			resp[NetworkServer.DAT] = (byte)adr;
			resp[NetworkServer.DAT + 1] = (byte)(adr >> 8);
			state = STATE_BNKLD;
			break;
		case STATE_BNKLD: // send BNK records
			// don't land here unless comlen > 0 (1+ records)
			resp[NetworkServer.FNC] = (byte)0x03;
			resp[NetworkServer.SIZ] = (byte)(RECLEN-1);
			n = netBoot.nextBnkRec(resp, NetworkServer.DAT, RECLEN);
			if (n <= 0) {
				if (org0 != null) {
					state = STATE_ORG0DMA;
				} else {
					state = STATE_DONE;
				}
			}
			break;
		case STATE_ORG0DMA:
			adr = org0Adr;
			resp[NetworkServer.FNC] = (byte)0x02;
			resp[NetworkServer.SIZ] = (byte)(2-1);
			resp[NetworkServer.DAT] = (byte)adr;
			resp[NetworkServer.DAT + 1] = (byte)(adr >> 8);
			adr = netBoot.entryAddr();
			org0[3] = (byte)adr;
			org0[4] = (byte)(adr >> 8);
			curdma = 0;
			// TODO: handle non-multiple of RECLEN...
			reccnt = org0.length / RECLEN;
			state = STATE_ORG0LD;
			break;
		case STATE_ORG0LD:
			resp[NetworkServer.FNC] = (byte)0x03;
			resp[NetworkServer.SIZ] = (byte)(RECLEN-1);
			System.arraycopy(org0, curdma, resp, NetworkServer.DAT, RECLEN);
			curdma += RECLEN;
			--reccnt;
			if (reccnt <= 0) {
				state = STATE_DONE;
			}
			break;
		case STATE_DONE: // done with boot.
			if (org0 != null) {
				adr = org0Adr;
			} else {
				adr = netBoot.entryAddr();
			}
			resp[NetworkServer.FNC] = (byte)0x04;
			resp[NetworkServer.SIZ] = (byte)(2-1);
			resp[NetworkServer.DAT] = (byte)adr;
			resp[NetworkServer.DAT + 1] = (byte)(adr >> 8);
			state = -1;
			break;
		default:	// error
			state = -1;
			return error(msg);
		}
		// if reach here, send a message...
		resp[NetworkServer.FMT] = (byte)0xb1;
		resp[NetworkServer.DID] = msg[NetworkServer.SID];
		resp[NetworkServer.SID] = msg[NetworkServer.DID];
		return resp;
	}
}
