// Copyright (c) 2018 Douglas Miller <durgadas311@gmail.com>

import java.util.Vector;
import java.util.Properties;

// To simplify internal interfaces, treat as single set of bits
// no matter how many ports.
//
// H8:
//      |  7  |  6  |  5  |  4  |  3  |  2  |  1  |  0  |
//      |-----|-----|-----|-----|-----|-----|-----|-----|
// F0H  |beep |int1 | mon |int2 |  D  |  C  |  B  |  A  | (LSbyte)
//      |-----|-----|-----|-----|-----|-----|-----|-----|
// F1H  | dp  |  f  |  e  |  d  |  c  |  b  |  a  |  g  |
//      |-----|-----|-----|-----|-----|-----|-----|-----|
// F2H* | io1 | ss  |org0 | sp1 | a16 | sp0 |  *  |  *  | (MSbyte)
//      |-----|-----|-----|-----|-----|-----|-----|-----|
// F3H* |     |     |     |     |     |     | 2ms |     |
//      |-----|-----|-----|-----|-----|-----|-----|-----|
//
// * = 2mS clock, single-step duplicated on Norberto's Z80 CPU board


public class NorbysGpp implements H8Fpp {
	// keyboard switch states are maintained in low 16 bits of 'gpi'
	private int base;
	private int nports;
	private int gpi;
	private int gpo;
	private boolean change;
	private H89 sys;
	private int[] src;
	private Vector<GppListener> notif;
	private FPListener fpl;
	static final int h8_2mSBit = (1 << 6);
	static final int nb_2mSBit = (1 << 17);
	static final int h8_SSBit = (1 << 4);
	static final int nb_SSBit = (1 << 16);	// not working yet!
	static final int aux2mSBit = (1 << 25);
	static final int org0Bit_c = (1 << 21);
	static final int mSClockBit = (h8_2mSBit | nb_2mSBit);
	//static final int sStepBit = (h8_SSBit | nb_SSBit);
	static final int sStepBit = (h8_SSBit); // only one functional?
	static final int sideSelBit_c = (1 << 22);
	static final int speed0Bit_c = (1 << 18);
	static final int speed1Bit_c = (1 << 20);
	static final int speedBits_c = (speed0Bit_c | speed1Bit_c);
	static final int speed2MHz_c = 0;
	static final int speed4MHz_c = (speed1Bit_c);
	static final int speed8MHz_c = (speed0Bit_c);
	static final int speed10MHz_c = (speed0Bit_c | speed1Bit_c);
	private int maxclk = 10240000;
	private boolean h89int1 = true;
	private boolean h89ena2 = false; // mod to fix dueling 2mS clocks
	private boolean h8nofp = false;

	public NorbysGpp(Properties props, H89 sys) {
		this.sys = sys;
		src = new int[4];
		src[0] = sys.registerINT(1);	// H8 i.e. port 360Q
		src[2] = sys.registerINT(1);	// H89 i.e. port 362Q
		gpi = 0;
		notif = new Vector<GppListener>();
		String s = props.getProperty("sw1");
		if (s != null) {
			gpi |= (Integer.valueOf(s, 2) & 0xff) << 16;
		}
		s = props.getProperty("h8nofp");
		h8nofp = (s != null);
		s = props.getProperty("h89int1");
		if (s != null) {
			h89int1 = (Integer.valueOf(s, 2) != 0);
		}
		// New CPU boards have this mod by default, cannot be disabled.
		h89ena2 = (H89.isZ180(props) || H89.isNorbyFeat(props, "ROMX"));
		s = props.getProperty("h89ena2");
		if (!h89ena2 && s != null) {
			h89ena2 = (Integer.valueOf(s, 2) != 0);
		}
		s = props.getProperty("h8maxclk");
		if (s != null) {
			// Either "10" or "16"
			if (s.equals("16")) {
				maxclk = 16384000;
			}
		}
		if (h8nofp) {
			base = 0xf2;
			nports = 1;
		} else {
			base = 0xf0;
			nports = 3;
		}
		if (h89ena2) {
			++nports;
		}
		// Norberto's new Z80 board, with "full" gpp 0f2h...
		// Either bit enables 2mS intr, both must be off to disable.
		// If single-step is supported, there are two bits also.
		reset();
	}

	// For GUI, Keyboard switches
	public void setKey(int kn) {
		kn &= 0x0000ff;
		kn ^= 0x0000ff;
		gpi &= ~0x0000ff;
		gpi |= kn;
//System.err.format("keypad %06x\n", gpi);
	}

	public int getOrg0Bit() { return org0Bit_c; }
	public int get2mSClockBit() { return mSClockBit; }
	public int getSStepBit() { return sStepBit; }
	public boolean isSStepEna() { return ((gpo ^ h8_SSBit) & sStepBit) != 0; }
	public int getSideSelBit() { return sideSelBit_c; }
	public int getMEM1Bit() { return (1 << 19); }
	public int getIO1Bit() { return (1 << 23); }

	/// IODevice interface ///

	public void reset() {
		gpi |= 0x0000ff;
		gpo = 0x0000df;	// databus inverted, tricky
		change = false;
		sys.lowerINT(1, src[0]);
		sys.lowerINT(1, src[2]);
		sys.setSpeed(2048000);	// not handled by notify()
		notify(0xffffff); // might contradict device resets?
	}

	public int getBaseAddress() {
		return base;
	}
	public int getNumPorts() {
		return nports;
	}
	public String getDeviceName() { return null; }

	public int in(int port) {
		int value = gpi;
		int off = port - 0xf0; // do not use 'base'
		value >>= (off * 8);
		value &= 0xff;
		return value;
	}

	public void out(int port, int value) {
//System.err.format("Norbys gpio out %02x => %02x\n", value, port);
		int off = port - 0xf0; // do not use 'base'

		value &= 0xff;
		if (off < 2) {
			// TODO: only mark 'change' if display bits change
			change = true;
		}
		int mask = (0xff << (off * 8));
		value <<= (off * 8);
		
		if ((mSClockBit & mask) != 0) {
			sys.lowerINT(1, src[off]);
		}
		int nv = gpo;
		nv &= ~mask;
		nv |= value;
		newValue(nv);
	}

	private void newValue(int nv) {
		int diff = gpo ^ nv;
		gpo = nv;
		if ((diff & speedBits_c) != 0) {
			int spd = 2048000;
			int sp = (gpo & speedBits_c);
			if (sp == speed10MHz_c) {	// i.e. MAX
				spd = maxclk;
			} else if (sp == speed8MHz_c) {
				spd = 8192000;
			} else if (sp == speed4MHz_c) {
				spd = 4096000;
			}
			sys.setSpeed(spd);
		}
		notify(diff);	// Assume Front Panel is a listener...
	}

	private void notify(int diff) {
		if (diff == 0) {
			return;
		}
		for (GppListener lstn : notif) {
			if ((diff & lstn.interestedBits()) != 0) {
				lstn.gppNewValue(gpo);
			}
		}
	}

	public void addGppListener(GppListener lstn) {
		notif.add(lstn);
		lstn.gppNewValue(gpo);
	}

	public void setFPListener(FPListener lstn) {
		fpl = lstn;
	}

	public String dumpDebug() {
		String ret = new String();
		ret += "GP IN:";
		if (h89ena2) {
			ret += String.format(" %02x", (gpi >> 24) & 0xff);
		}
		ret += String.format(" %02x %02x %02x\n",
					(gpi >> 16) & 0xff,
					(gpi >> 8) & 0xff,
					gpi & 0xff);
		ret += "GP OUT:";
		if (h89ena2) {
			ret += String.format(" %02x", (gpo >> 24) & 0xff);
		}
		ret += String.format(" %02x %02x %02x\n",
					(gpo >> 16) & 0xff,
					(gpo >> 8) & 0xff,
					gpo & 0xff);
		ret += String.format("2mS Clock: %s %s\n",
				isH8int1(), isH89int1());
		ret += String.format("ORG0: %s\n", ((gpo & org0Bit_c) != 0));
		return ret;
	}

	private boolean isH8int1() {
		return (!h8nofp && (gpo & h8_2mSBit) != 0);
	}
	private boolean isH89int1() {
		return (h89int1 && (gpo & nb_2mSBit) != 0 &&
			(!h89ena2 || (gpo & aux2mSBit) != 0));
	}

	public void trigger2ms() {
		// actually, at -0.5 mS...
		if (fpl != null) {
			fpl.update();
			if (change) {
				change = false;
				fpl.timeout(gpo);
			}
		}
		newValue((gpo & ~0x0020) | 0x00ff00);
		if (isH8int1()) {
			sys.raiseINT(1, src[0]);
		}
		if (isH89int1()) {
			sys.raiseINT(1, src[2]);
		}
	}
}
