// Copyright 2017 Douglas Miller <durgadas311@gmail.com>

import java.util.Arrays;
import java.util.Date;
import java.util.Map;
import java.util.HashMap;
import java.util.Properties;
import java.awt.event.*;
import java.text.SimpleDateFormat;
import java.text.ParsePosition;

public class RTC72421 implements IODevice, ActionListener {
	private int baseAdr;
	private int[] regs;
	private long off = 0;	// in milliseconds.
	private int intr;
	private boolean hold;
	private boolean dirty;
	private javax.swing.Timer[] timers;

	public static SimpleDateFormat timestamp =
		new java.text.SimpleDateFormat("yyyy MM dd EEE HH:mm:ss");
	private static Map<String, Integer> wdays;
	static {
		wdays = new HashMap<String, Integer>();
		wdays.put("Sun", 0);
		wdays.put("Mon", 1);
		wdays.put("Tue", 2);
		wdays.put("Wed", 3);
		wdays.put("Thu", 4);
		wdays.put("Fri", 5);
		wdays.put("Sat", 6);
	}
	private static String[] xdays = new String[]{
		"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat", "Sat"
	};
	private static byte[] wmsk = new byte[]{
		(byte)0b1111, (byte)0b0111, (byte)0b1111, (byte)0b0111,
		(byte)0b1111, (byte)0b0111, // includes AM/PM bit
		(byte)0b1111, (byte)0b0011, (byte)0b1111, (byte)0b0001,
		(byte)0b1111, (byte)0b1111, (byte)0b0111,
		// control regs:
		(byte)0b1101, (byte)0b1111, (byte)0b1111
	};

	public RTC72421(Properties pros, int base) {
		baseAdr = base;
		regs = new int[16]; // these are 4-bit registers
		timers = new javax.swing.Timer[8];
		Arrays.fill(regs, 0);
		off = 0;
		reset();
		// pretend time already set...
		getTime();
		System.err.format("RTC72421 device at port %02x\n", baseAdr);
	}

	private void getTime() {
		long time = off + new Date().getTime();
		Date dt = new Date(time);
		String tod = timestamp.format(dt);
		regs[11] = (tod.charAt(2) & 0x0f);	// Y10
		regs[10] = (tod.charAt(3) & 0x0f);	// Y1
		regs[9] = (tod.charAt(5) & 0x0f);	// MO10
		regs[8] = (tod.charAt(6) & 0x0f);	// MO10
		regs[7] = (tod.charAt(8) & 0x0f);	// D10
		regs[6] = (tod.charAt(9) & 0x0f);	// D10
		Integer w = wdays.get(tod.substring(11, 14));
		if (w == null) {
			regs[12] = 0;
		} else {
			regs[12] = w;
		}
		regs[5] = (tod.charAt(15) & 0x0f);	// H10
		regs[4] = (tod.charAt(16) & 0x0f);	// H1
		if ((regs[15] & 0b0100) == 0) {
			int hr = regs[5] * 10 + regs[4];
			int pm = 0;
			if (hr == 0) {
				hr = 12;
				pm = 0b0100;
			} else if (hr > 12) {
				hr -= 12;
				pm = 0b0100;
			}
			regs[5] = pm | (hr / 10);
			regs[4] = (hr % 10);
		}
		regs[3] = (tod.charAt(18) & 0x0f);	// MI10
		regs[2] = (tod.charAt(19) & 0x0f);	// MI1
		regs[1] = (tod.charAt(21) & 0x0f);	// S10
		regs[0]	= (tod.charAt(22) & 0x0f);	// S1
		dirty = false;
	}

	private void setTime() {
		int hr = (regs[5] * 10) + regs[4];
		if ((regs[15] & 0b0100) == 0) {
			hr &= 0x1f;
			if ((regs[5] & 0x40) != 0) {
				hr += 12;
			}
		}
		Date now = new Date();
		String t = timestamp.format(now);
		String tod = t.substring(0, 2) +
			String.format("%d%d %d%d %d%d %s %02d:%d%d:%d%d",
				regs[11], regs[10], regs[9], regs[8],
				regs[7], regs[6], xdays[regs[12] & 7],
				hr, regs[3], regs[2],
				regs[1], regs[0]);
		Date dt = timestamp.parse(tod, new ParsePosition(0));
		off = dt.getTime() - now.getTime(); // milliseconds
		dirty = false;
	}

	private void chkIntr() {
		// TODO: Not how we do things?
	}

	static final int[] times = new int[] {
		// TODO: interrupt handling
		-1,		// 0: comparator - custom timeout, one-shot(?)
	};

	private void setIntr() {
		// TODO: interrupt handling
	}

	private int getAddr(int port) {
		return (port - baseAdr) & 0x0f;
	}

	public void reset() {
		// TODO: does anything get reset?
		hold = false;
		dirty = false;
	}

	public int getBaseAddress() { return baseAdr; }
	public int getNumPorts() { return 16; }
	public int in(int port) {
		int val = 0;
		int adr = getAddr(port);
		val = regs[adr] | 0xf0;	// simulate floating data lines D4-D7
		// TODO: any post-processing?
		return val;
	}
	public void out(int port, int value) {
		int adr = getAddr(port);
		regs[adr] = value & wmsk[adr];
		switch (adr) {
		case 13:	// HOLD bit, others
			boolean h = ((value & 1) != 0);
			if (h != hold) {
				if (h) {
					getTime();
				} else if (dirty) {
					setTime();
				}
				hold = h;
			}
			break;
		case 14:	// Interrupt Control
			// TODO: implement this
			break;
		case 15:	// Counters reset
			// RESET only affects < 1 second. no-op for us
			// STOP is also not applicable.
			break;
		default:
			dirty = true;
			break;
		}
	}

	public void actionPerformed(ActionEvent e) {
		// TODO: interrupt handling
	}

	public String getDeviceName() {
		return "RTC72421";
	}
	public String dumpDebug() {
		String str = String.format("port %02x\n", baseAdr);
		str += String.format("%x %x %x %x %x %x %x %x %x %x %x %x %x\n",
			regs[0], regs[1], regs[2], regs[3],
			regs[4], regs[5],
			regs[6], regs[7],
			regs[8], regs[9], regs[10], regs[11],
			regs[12]);
		str += String.format("(%s %d%d/%d%d/%d%d %c%d%d:%d%d:%d%d)\n",
			xdays[regs[12] & 7],
			regs[9], regs[8], regs[7], regs[6], regs[11], regs[10],
			(regs[5] & 0x04) != 0 ? '*' : ' ',
			regs[5] & 0x03, regs[4], regs[3], regs[2], regs[1], regs[0]);
		str += String.format("%02x %02x %02x delta:%d mS\n",
			regs[13], regs[14], regs[15], off);
		return str;
	}
}
