// Copyright (c) 2016 Douglas Miller <durgadas311@gmail.com>

import java.util.Arrays;
import java.util.Vector;
import java.util.Properties;

public class Z37 extends WD1797
		implements DiskController, InterruptController, WD1797Controller {
	// TODO: These need to come from CPU (Z80)...
	static final int EI = 0xfb;
	static final int RST7 = 0xff;

	static final int BasePort_c = 0x78;
	static final int Z37_NumPorts_c = 4;
	static final int ControlPort_Offset_c = 0;
	static final int MuxPort_Offset_c = 1;
	static final int Wd1797_Offset_c = 2; // 4 addresses multiplexed into 2.
	static final int numDisks_c = 4;
	static final int Z37_IntrqIntr_c = 4;
	static final int Z37_DrqIntr_c = 7;

	static final int ctrl_EnableIntReq_c = 0x01;
	static final int ctrl_EnableDReq_c = 0x02;
	static final int ctrl_SetMFMRecordingN_c = 0x04;
	static final int ctrl_MotorOn_c = 0x08;
	static final int ctrl_DriveSel_c = 0xf0; // NOTE! bitmap not index!
	static final int ctrl_DriveSelShift_c = 4;

	private String Z37_Name = "Z37";
	private GenericFloppyDrive[] drives_m;
	private GenericFloppyDrive curDrive = null;
	private LED[] leds_m;
	private int BasePort = BasePort_c;
	private int srcDrq;
	private int srcIntrq;
	static final int blockedInts = 0x6f;
	private Interruptor intr;

	private int controlReg_m;
	private int curDriveIndex; // convenience decode of bitmap
	private int muxReg_m;

	public Z37(Properties props, String slot, LEDHandler lh, Interruptor intr) {
		super(BasePort_c, intr);  // FDC only sees modified port addr...
		super.setController(this);
		this.intr = intr;
		srcDrq = intr.registerINT(Z37_IntrqIntr_c);
		srcIntrq = intr.registerINT(Z37_DrqIntr_c);
		intr.addIntrController(this);
		drives_m = new GenericFloppyDrive[numDisks_c];
		leds_m = new LED[numDisks_c];
		String s;
		Arrays.fill(drives_m, null);
		controlReg_m = 0;
		muxReg_m = 0;

		s = props.getProperty(slot);
		String[] ss = s.split("\\s");
		Z37_Name = ss[0];
		String pfx = Z37_Name.toLowerCase();

		// NOTE: installation document dictates port 0x78 and slot 504,
		// So this shouldn't really be configurable. But the board
		// could be forced to other ports even if the software does
		// not support it. The following is based on 444-61C (MMS).

		s = props.getProperty(pfx + "_j1"); // "A" or "B"
		if (s == null) {
			s = props.getProperty(pfx + "_p1"); // "A" or "B"
		}
		if (s != null && s.equalsIgnoreCase("b")) {
			BasePort ^= 0x04; // 78 -> 7C
		}
		// "J2" selects which slot:
		// "A" = 504/505: use "IOCASS" = 78-7F (J1 selects 78 or 7C)
		// "B" = 506: use "IOFLPY" = 7C-7F
		s = props.getProperty(pfx + "_j2"); // "A" or "B"
		if (s == null) {
			s = props.getProperty(pfx + "_p2"); // "A" or "B"
		}
		if (s != null && s.equalsIgnoreCase("b")) {
			BasePort = 0x7c;
		}
		super.basePort_m = BasePort;

		// First identify what drives are installed.
		GenericFloppyDrive drv;
		for (int x = 0; x < numDisks_c; ++x) {
			String prop = String.format(pfx + "_drive%d", x + 1);
			s = props.getProperty(prop);
			if (s != null) {
				drv = installDrive(x, s, lh);
			}
		}

		// Next identify what diskettes are pre-inserted.
		for (int x = 0; x < numDisks_c; ++x) {
			String prop = String.format(pfx + "_disk%d", x + 1);
			s = props.getProperty(prop);
			if (s != null) {
				drv = getDrive(x);
				if (drv != null) {
					drv.insertMedia(s.split("\\s"));
				}
			}
		}
		System.err.format("%s at port %02x\n", Z37_Name, BasePort);
		reset();
	}

	private GenericFloppyDrive installDrive(int x, String s, LEDHandler lh) {
		String[] ss = s.split("\\s", 2);
		String n = getDriveName(x);
		if (ss.length >= 2) {
			n = ss[1];
		}
		GenericFloppyDrive drv = GenericFloppyDrive.getInstance(ss[0], n, lh);
		if (!connectDrive(x, drv)) {
			return null;
		}
		leds_m[x] = drv.getLED();
		return drv;
	}

	public int getBaseAddress() { return BasePort; }
	public int getNumPorts() { return Z37_NumPorts_c; }

	public void setCurLed(boolean on) {
		// This could be null
		if (curDriveIndex >= drives_m.length) {
			return;
		}
		if (leds_m[curDriveIndex] != null) {
			leds_m[curDriveIndex].set(on);
		}
	}

	public GenericFloppyDrive getCurDrive() {
		// This could be null
		if (curDriveIndex >= drives_m.length) {
			return null;
		}
		return drives_m[curDriveIndex];
	}

	public int getClockPeriod() {
		return 1000; // always 5" drive clock...
	}

	public Vector<GenericRemovableDrive> getDiskDrives() {
		Vector<GenericRemovableDrive> drives = new Vector<GenericRemovableDrive>();
		for (int x = 0; x < numDisks_c; ++x) {
			drives.add(drives_m[x]);	// might be null, but must preserve number.
		}
		return drives;
	}

	public String getDriveName(int index) {
		if (index < 0 || index >= numDisks_c) {
			return null;
		}
		String str = String.format("%s-%d", Z37_Name, index + 1);
		return str;
	}

	public String getDeviceName() { return Z37_Name; }

	public GenericRemovableDrive findDrive(String ident) {
		for (int x = 0; x < numDisks_c; ++x) {
			if (drives_m[x] == null) {
				continue;
			}
			if (ident.equals(drives_m[x].getDriveName())) {
				return drives_m[x];
			}
		}
		return null;
	}

	public int readDataBus() {
		int opCode = -1;
		// The Z37 controller assumes it is the highest priority interrupt,
		// so if we have a DRQ interrupt asserted then we take control.
		if (drqAllowed() && drqRaised_m) {
			// TODO: generate op-codes without knowing Z80CPU?
			opCode = (drqRaised_m ? EI : RST7);
		}
		return opCode;
	}

	public void destroy() {
		for (int x = 0; x < numDisks_c; ++x) {
			if (drives_m[x] != null) {
				drives_m[x].insertMedia(null);
			}
		}
	}

	private void setControlReg(int val) {
		int diff = controlReg_m ^ val;
		controlReg_m = val;
		if ((controlReg_m & ctrl_EnableIntReq_c) != 0) {
			// Interrupts now unmasked, so raise as appropriate.
			if (intrqRaised_m) {
				intr.raiseINT(Z37_IntrqIntr_c, srcIntrq);
			}
		}
		if ((diff & ctrl_EnableDReq_c) != 0) {
			boolean drqEn = ((controlReg_m & ctrl_EnableDReq_c) != 0);
			if (drqEn) {
				intr.blockInts(blockedInts);
				if (drqRaised_m) {
					intr.raiseINT(Z37_DrqIntr_c, srcDrq);
				}
			} else {
				intr.unblockInts(blockedInts);
			}
		}
		if ((diff & ctrl_DriveSel_c) != 0) {
			setCurLed(false);
			int dsx = (controlReg_m & ctrl_DriveSel_c) >> ctrl_DriveSelShift_c;
			curDriveIndex = Integer.numberOfTrailingZeros(dsx);
			if (curDriveIndex < drives_m.length) {
				curDrive = drives_m[curDriveIndex];
				if (curDrive != null) {
					curDrive.motor((controlReg_m & ctrl_MotorOn_c) != 0);
				}
			} else {
				curDrive = null;
			}
			setCurLed(true);
		} else if (curDrive != null) {
			if ((diff & ctrl_MotorOn_c) != 0) {
				curDrive.motor((controlReg_m & ctrl_MotorOn_c) != 0);
			}
		}
	}

	public void reset() {
		super.reset();
		setControlReg(0);
		muxReg_m = 0;
		intr.lowerINT(Z37_DrqIntr_c, srcDrq);
		intr.lowerINT(Z37_IntrqIntr_c, srcIntrq);
	}

	public int in(int addr) {
		int offset = addr - BasePort;
		int val = 0;
		// case ControlPort_Offset_c: NOT READABLE
		if (offset >= Wd1797_Offset_c) {
			// kludge port addresses...
			int wdadr = addr;
			if (muxReg_m == (addr & 1)) {
				wdadr ^= 0x02;
			}
			val = super.in(wdadr);
			if (wdadr - BasePort == StatusPort_Offset_c) {
				// H/Z37 forces READY
				val &= ~stat_NotReady_c;
			}
		} else {
			System.err.format("Z37::in(Unknown - 0x%02x)\n", addr);
		}
		return val;
	}

	public void out(int addr, int val) {
		int offset = addr - BasePort;
		if (offset >= Wd1797_Offset_c) {
			// kludge port addresses...
			int wdadr = addr;
			if (muxReg_m == (addr & 1)) {
				wdadr ^= 0x02;
			}
			super.out(wdadr, val);
		} else if (offset == ControlPort_Offset_c) {
			setControlReg(val);
		} else if (offset == MuxPort_Offset_c) {
			// TODO: any other bits used?
			muxReg_m = val & 1;
		} else {
			System.err.format("Z37::out(Unknown - 0x%02x): %d\n", addr, val);
		}
	}

	public GenericFloppyDrive getDrive(int unitNum) {
		if (unitNum < numDisks_c) {
			return drives_m[unitNum];
		}
		return null;
	}

	public boolean connectDrive(int unitNum, GenericFloppyDrive drive) {
		if (unitNum >= numDisks_c) {
			System.err.format("Z37 Invalid unit number (%d)\n", unitNum);
			return false;
		}
		if (drives_m[unitNum] != null) {
			System.err.format("Z37 %d: drive already connect\n", unitNum);
			return false;
		}
		if (drive.getMediaSize() != 5) {
			System.err.format("Z37 %d: drive incompatible\n", unitNum);
			return false;
		}
		drives_m[unitNum] = drive;
		return true;
	}

	public boolean removeDrive(int unitNum) {
		return false;
	}

	public void raisedIntrq() {
		if (intrqAllowed()) {
			intr.raiseINT(Z37_IntrqIntr_c, srcIntrq);
			return;
		}
	}

	public void raisedDrq() {
		if (drqAllowed()) {
			intr.raiseINT(Z37_DrqIntr_c, srcDrq);
		}
	}

	public void loweredIntrq() {
		intr.lowerINT(Z37_IntrqIntr_c, srcIntrq);
	}

	public void loweredDrq() {
		intr.lowerINT(Z37_DrqIntr_c, srcDrq);
	}

	public void loadedHead(boolean load) {
		// 5" drives don't have head-load...
		if (!load) {
			statusReg_m &= ~stat_HeadLoaded_c;
		} else {
			statusReg_m |= stat_HeadLoaded_c;
		}
	}

	private boolean intrqAllowed() {
		return (controlReg_m & ctrl_EnableIntReq_c) != 0;
	}
	private boolean drqAllowed() {
		return (controlReg_m & ctrl_EnableDReq_c) != 0;
	}
	public boolean doubleDensity() {
		return (controlReg_m & ctrl_SetMFMRecordingN_c) != 0;
	}

	public String dumpDebug() {
		String ret = String.format(
			"CTRL=%02x MUX=%02x\n" +
			"FDC-STS=%02x FDC-CMD=%02x\n" +
			"FDC-TRK=%d FDC-SEC=%d FDC-DAT=%02x\n" +
			"DRQ=%s INTRQ=%s\n",
			controlReg_m, muxReg_m,
			statusReg_m, cmdReg_m,
			trackReg_m, sectorReg_m, dataReg_m,
			Boolean.toString(drqRaised_m), Boolean.toString(intrqRaised_m));
		return ret;
	}
}
